# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING
from unittest import mock

import boto3
import pytest
from boto3 import client
from moto import mock_aws

from airflow.exceptions import AirflowException, TaskDeferred
from airflow.providers.amazon.aws.hooks.glue import GlueDataQualityHook, GlueJobHook
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.links.glue import GlueJobRunDetailsLink
from airflow.providers.amazon.aws.operators.glue import (
    GlueDataQualityOperator,
    GlueDataQualityRuleRecommendationRunOperator,
    GlueDataQualityRuleSetEvaluationRunOperator,
    GlueJobOperator,
)

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

if TYPE_CHECKING:
    from airflow.models import TaskInstance
    from airflow.providers.amazon.aws.hooks.base_aws import BaseAwsConnection

TASK_ID = "test_glue_operator"
DAG_ID = "test_dag_id"
JOB_NAME = "test_job_name/with_slash"
JOB_RUN_ID = "11111"


class TestGlueJobOperator:
    @pytest.mark.db_test
    def test_render_template(self, create_task_instance_of_operator, session):
        ti: TaskInstance = create_task_instance_of_operator(
            GlueJobOperator,
            dag_id=DAG_ID,
            task_id=TASK_ID,
            script_location="{{ dag.dag_id }}",
            script_args="{{ dag.dag_id }}",
            create_job_kwargs="{{ dag.dag_id }}",
            iam_role_name="{{ dag.dag_id }}",
            iam_role_arn="{{ dag.dag_id }}",
            s3_bucket="{{ dag.dag_id }}",
            job_name="{{ dag.dag_id }}",
        )
        session.add(ti)
        session.commit()
        rendered_template: GlueJobOperator = ti.render_templates()

        assert rendered_template.script_location == DAG_ID
        assert rendered_template.script_args == DAG_ID
        assert rendered_template.create_job_kwargs == DAG_ID
        assert rendered_template.iam_role_name == DAG_ID
        assert rendered_template.iam_role_arn == DAG_ID
        assert rendered_template.s3_bucket == DAG_ID
        assert rendered_template.job_name == DAG_ID

    @pytest.mark.parametrize(
        "script_location",
        [
            "s3://glue-examples/glue-scripts/sample_aws_glue_job.py",
            "/glue-examples/glue-scripts/sample_aws_glue_job.py",
        ],
    )
    @mock.patch.object(GlueJobHook, "print_job_logs")
    @mock.patch.object(GlueJobHook, "get_job_state")
    @mock.patch.object(GlueJobHook, "initialize_job")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(S3Hook, "load_file")
    def test_execute_without_failure(
        self,
        mock_load_file,
        mock_get_conn,
        mock_initialize_job,
        mock_get_job_state,
        mock_print_job_logs,
        script_location,
    ):
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location=script_location,
            aws_conn_id="aws_default",
            region_name="us-west-2",
            s3_bucket="some_bucket",
            iam_role_name="my_test_role",
        )
        mock_initialize_job.return_value = {"JobRunState": "RUNNING", "JobRunId": JOB_RUN_ID}
        mock_get_job_state.return_value = "SUCCEEDED"

        glue.execute(mock.MagicMock())

        mock_initialize_job.assert_called_once_with({}, {})
        mock_print_job_logs.assert_not_called()
        assert glue.job_name == JOB_NAME

    @mock.patch.object(GlueJobHook, "initialize_job")
    @mock.patch.object(GlueJobHook, "get_conn")
    def test_role_arn_execute_deferrable(self, _, mock_initialize_job):
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location="s3://folder/file",
            aws_conn_id="aws_default",
            region_name="us-west-2",
            s3_bucket="some_bucket",
            iam_role_arn="test_role",
            deferrable=True,
        )
        mock_initialize_job.return_value = {"JobRunState": "RUNNING", "JobRunId": JOB_RUN_ID}

        with pytest.raises(TaskDeferred) as defer:
            glue.execute(mock.MagicMock())

        assert defer.value.trigger.job_name == JOB_NAME
        assert defer.value.trigger.run_id == JOB_RUN_ID

    @mock.patch.object(GlueJobHook, "initialize_job")
    @mock.patch.object(GlueJobHook, "get_conn")
    def test_execute_deferrable(self, _, mock_initialize_job):
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location="s3://folder/file",
            aws_conn_id="aws_default",
            region_name="us-west-2",
            s3_bucket="some_bucket",
            iam_role_name="my_test_role",
            deferrable=True,
        )
        mock_initialize_job.return_value = {"JobRunState": "RUNNING", "JobRunId": JOB_RUN_ID}

        with pytest.raises(TaskDeferred) as defer:
            glue.execute(mock.MagicMock())

        assert defer.value.trigger.job_name == JOB_NAME
        assert defer.value.trigger.run_id == JOB_RUN_ID

    @mock.patch.object(GlueJobHook, "print_job_logs")
    @mock.patch.object(GlueJobHook, "get_job_state")
    @mock.patch.object(GlueJobHook, "initialize_job")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(S3Hook, "load_file")
    def test_execute_with_verbose_logging(
        self, mock_load_file, mock_get_conn, mock_initialize_job, mock_get_job_state, mock_print_job_logs
    ):
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location="s3_uri",
            s3_bucket="bucket_name",
            iam_role_name="role_arn",
            verbose=True,
        )
        mock_initialize_job.return_value = {"JobRunState": "RUNNING", "JobRunId": JOB_RUN_ID}
        mock_get_job_state.return_value = "SUCCEEDED"

        glue.execute(mock.MagicMock())

        mock_initialize_job.assert_called_once_with({}, {})
        mock_print_job_logs.assert_called_once_with(
            job_name=JOB_NAME, run_id=JOB_RUN_ID, continuation_tokens=mock.ANY
        )
        assert glue.job_name == JOB_NAME

    @mock.patch.object(GlueJobHook, "print_job_logs")
    @mock.patch.object(GlueJobHook, "get_job_state")
    @mock.patch.object(GlueJobHook, "initialize_job")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(S3Hook, "load_file")
    def test_execute_without_verbose_logging(
        self, mock_load_file, mock_get_conn, mock_initialize_job, mock_get_job_state, mock_print_job_logs
    ):
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location="s3_uri",
            s3_bucket="bucket_name",
            iam_role_name="role_arn",
            verbose=False,
        )
        mock_initialize_job.return_value = {"JobRunState": "RUNNING", "JobRunId": JOB_RUN_ID}
        mock_get_job_state.return_value = "SUCCEEDED"

        glue.execute(mock.MagicMock())

        mock_initialize_job.assert_called_once_with({}, {})
        mock_print_job_logs.assert_not_called()
        assert glue.job_name == JOB_NAME

    @mock.patch.object(GlueJobHook, "print_job_logs")
    @mock.patch.object(GlueJobHook, "job_completion")
    @mock.patch.object(GlueJobHook, "initialize_job")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(S3Hook, "load_file")
    def test_execute_without_waiting_for_completion(
        self, mock_load_file, mock_get_conn, mock_initialize_job, mock_job_completion, mock_print_job_logs
    ):
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location="s3://glue-examples/glue-scripts/sample_aws_glue_job.py",
            aws_conn_id="aws_default",
            region_name="us-west-2",
            s3_bucket="some_bucket",
            iam_role_name="my_test_role",
            wait_for_completion=False,
        )
        mock_initialize_job.return_value = {"JobRunState": "RUNNING", "JobRunId": JOB_RUN_ID}

        job_run_id = glue.execute(mock.MagicMock())

        mock_initialize_job.assert_called_once_with({}, {})
        mock_job_completion.assert_not_called()
        mock_print_job_logs.assert_not_called()
        assert glue.job_name == JOB_NAME
        assert job_run_id == JOB_RUN_ID

    @mock.patch.object(GlueJobHook, "print_job_logs")
    @mock.patch.object(GlueJobHook, "get_job_state")
    @mock.patch.object(GlueJobHook, "initialize_job")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(S3Hook, "load_file")
    def test_log_correct_url(
        self, mock_load_file, mock_get_conn, mock_initialize_job, mock_get_job_state, mock_print_job_logs
    ):
        region = "us-west-2"
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location="s3://glue-examples/glue-scripts/sample_aws_glue_job.py",
            aws_conn_id="aws_default",
            region_name=region,
            s3_bucket="some_bucket",
            iam_role_name="my_test_role",
        )
        mock_initialize_job.return_value = {"JobRunState": "RUNNING", "JobRunId": JOB_RUN_ID}
        mock_get_job_state.return_value = "SUCCEEDED"

        aws_domain = GlueJobRunDetailsLink.get_aws_domain("aws")
        glue_job_run_url = (
            f"https://console.{aws_domain}/gluestudio/home?region="
            f"{region}#/job/test_job_name%2Fwith_slash/run/{JOB_RUN_ID}"
        )

        with mock.patch.object(glue.log, "info") as mock_log_info:
            job_run_id = glue.execute(mock.MagicMock())
            assert job_run_id == JOB_RUN_ID

        mock_log_info.assert_any_call("You can monitor this Glue Job run at: %s", glue_job_run_url)

    @mock.patch.object(GlueJobHook, "conn")
    @mock.patch.object(GlueJobHook, "get_conn")
    def test_killed_without_stop_job_run_on_kill(
        self,
        _,
        mock_glue_hook,
    ):
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location="s3://folder/file",
            aws_conn_id="aws_default",
            region_name="us-west-2",
            s3_bucket="some_bucket",
            iam_role_name="my_test_role",
        )
        glue.on_kill()
        mock_glue_hook.batch_stop_job_run.assert_not_called()

    @mock.patch.object(GlueJobHook, "conn")
    @mock.patch.object(GlueJobHook, "get_conn")
    def test_killed_with_stop_job_run_on_kill(
        self,
        _,
        mock_glue_hook,
    ):
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location="s3://folder/file",
            aws_conn_id="aws_default",
            region_name="us-west-2",
            s3_bucket="some_bucket",
            iam_role_name="my_test_role",
            stop_job_run_on_kill=True,
        )
        glue._job_run_id = JOB_RUN_ID
        glue.on_kill()
        mock_glue_hook.batch_stop_job_run.assert_called_once_with(
            JobName=JOB_NAME,
            JobRunIds=[JOB_RUN_ID],
        )

    @mock.patch.object(GlueJobHook, "get_job_state")
    @mock.patch.object(GlueJobHook, "initialize_job")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(GlueJobHook, "conn")
    @mock.patch.object(S3Hook, "load_file")
    def test_replace_script_file(
        self, mock_load_file, mock_conn, mock_get_connection, mock_initialize_job, mock_get_job_state
    ):
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location="folder/file",
            s3_bucket="bucket_name",
            iam_role_name="role_arn",
            replace_script_file=True,
        )
        mock_initialize_job.return_value = {"JobRunState": "RUNNING", "JobRunId": JOB_RUN_ID}
        mock_get_job_state.return_value = "SUCCEEDED"
        glue.execute(mock.MagicMock())
        mock_load_file.assert_called_once_with(
            "folder/file", "artifacts/glue-scripts/file", bucket_name="bucket_name", replace=True
        )

        assert glue.s3_script_location == "s3://bucket_name/artifacts/glue-scripts/file"

    @mock.patch.object(GlueJobHook, "get_job_state")
    @mock.patch.object(GlueJobHook, "initialize_job")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(GlueJobHook, "conn")
    @mock.patch.object(S3Hook, "load_file")
    @mock.patch.object(GlueJobOperator, "upload_etl_script_to_s3")
    def test_upload_script_to_s3_no_upload(
        self,
        mock_upload,
        mock_load_file,
        mock_conn,
        mock_get_connection,
        mock_initialize_job,
        mock_get_job_state,
    ):
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location="s3://my_bucket/folder/file",
            s3_bucket="bucket_name",
            iam_role_name="role_arn",
            replace_script_file=True,
        )
        mock_initialize_job.return_value = {"JobRunState": "RUNNING", "JobRunId": JOB_RUN_ID}
        mock_get_job_state.return_value = "SUCCEEDED"
        glue.execute(mock.MagicMock())

        assert glue.s3_script_location == "s3://my_bucket/folder/file"
        mock_load_file.assert_not_called()
        mock_upload.assert_not_called()

    @mock.patch.object(GlueJobHook, "get_job_state")
    @mock.patch.object(GlueJobHook, "initialize_job")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(GlueJobHook, "conn")
    @mock.patch.object(S3Hook, "load_file")
    @mock.patch.object(GlueJobOperator, "upload_etl_script_to_s3")
    def test_no_script_file(
        self,
        mock_upload,
        mock_load_file,
        mock_conn,
        mock_get_connection,
        mock_initialize_job,
        mock_get_job_state,
    ):
        glue = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            iam_role_name="role_arn",
            replace_script_file=True,
        )

        mock_initialize_job.return_value = {"JobRunState": "RUNNING", "JobRunId": JOB_RUN_ID}
        mock_get_job_state.return_value = "SUCCEEDED"
        glue.execute(mock.MagicMock())

        assert glue.s3_script_location is None
        mock_upload.assert_not_called()

    def test_template_fields(self):
        operator = GlueJobOperator(
            task_id=TASK_ID,
            job_name=JOB_NAME,
            script_location="folder/file",
            s3_bucket="bucket_name",
            iam_role_name="role_arn",
            replace_script_file=True,
        )
        validate_template_fields(operator)

    def test_overwritten_conn_passed_to_hook(self):
        OVERWRITTEN_CONN = "new-conn-id"
        op = GlueJobOperator(
            task_id=TASK_ID,
            aws_conn_id=OVERWRITTEN_CONN,
            iam_role_name="role_arn",
            replace_script_file=True,
        )
        assert op.hook.aws_conn_id == OVERWRITTEN_CONN

    def test_default_conn_passed_to_hook(self):
        DEFAULT_CONN = "aws_default"
        op = GlueJobOperator(
            task_id=TASK_ID,
            iam_role_name="role_arn",
            replace_script_file=True,
        )
        assert op.hook.aws_conn_id == DEFAULT_CONN


class TestGlueDataQualityOperator:
    RULE_SET_NAME = "TestRuleSet"
    RULE_SET = 'Rules=[ColumnLength "review_id" = 15]'
    TARGET_TABLE = {"TableName": "TestTable", "DatabaseName": "TestDB"}

    @pytest.fixture
    def glue_data_quality_hook(self) -> Generator[GlueDataQualityHook, None, None]:
        with mock_aws():
            hook = GlueDataQualityHook(aws_conn_id="aws_default")
            yield hook

    def test_init(self):
        self.operator = GlueDataQualityOperator(
            task_id="create_data_quality_ruleset", name=self.RULE_SET_NAME, ruleset=self.RULE_SET
        )
        self.operator.defer = mock.MagicMock()

        assert self.operator.name == self.RULE_SET_NAME
        assert self.operator.ruleset == self.RULE_SET

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_execute_create_rule(self, glue_data_quality_mock_conn):
        self.operator = GlueDataQualityOperator(
            task_id="create_data_quality_ruleset",
            name=self.RULE_SET_NAME,
            ruleset=self.RULE_SET,
            description="create ruleset",
        )
        self.operator.defer = mock.MagicMock()

        self.operator.execute({})
        glue_data_quality_mock_conn.create_data_quality_ruleset.assert_called_once_with(
            Description="create ruleset",
            Name=self.RULE_SET_NAME,
            Ruleset=self.RULE_SET,
        )

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_execute_create_rule_should_fail_if_rule_already_exists(self, glue_data_quality_mock_conn):
        self.operator = GlueDataQualityOperator(
            task_id="create_data_quality_ruleset",
            name=self.RULE_SET_NAME,
            ruleset=self.RULE_SET,
            description="create ruleset",
        )
        self.operator.defer = mock.MagicMock()
        error_message = f"Another ruleset with the same name already exists: {self.RULE_SET_NAME}"

        err_response = {"Error": {"Code": "AlreadyExistsException", "Message": error_message}}

        exception = client("glue").exceptions.ClientError(err_response, "test")
        returned_exception = type(exception)

        glue_data_quality_mock_conn.exceptions.AlreadyExistsException = returned_exception
        glue_data_quality_mock_conn.create_data_quality_ruleset.side_effect = exception

        with pytest.raises(AirflowException, match=error_message):
            self.operator.execute({})

        glue_data_quality_mock_conn.create_data_quality_ruleset.assert_called_once_with(
            Description="create ruleset",
            Name=self.RULE_SET_NAME,
            Ruleset=self.RULE_SET,
        )

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_execute_update_rule(self, glue_data_quality_mock_conn):
        self.operator = GlueDataQualityOperator(
            task_id="update_data_quality_ruleset",
            name=self.RULE_SET_NAME,
            ruleset=self.RULE_SET,
            description="update ruleset",
            update_rule_set=True,
        )
        self.operator.defer = mock.MagicMock()

        self.operator.execute({})
        glue_data_quality_mock_conn.update_data_quality_ruleset.assert_called_once_with(
            Description="update ruleset", Name=self.RULE_SET_NAME, Ruleset=self.RULE_SET
        )

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_execute_update_rule_should_fail_if_rule_not_exists(self, glue_data_quality_mock_conn):
        self.operator = GlueDataQualityOperator(
            task_id="update_data_quality_ruleset",
            name=self.RULE_SET_NAME,
            ruleset=self.RULE_SET,
            description="update ruleset",
            update_rule_set=True,
        )
        self.operator.defer = mock.MagicMock()
        error_message = f"Cannot find Data Quality Ruleset in account 1234567 with name {self.RULE_SET_NAME}"

        err_response = {"Error": {"Code": "EntityNotFoundException", "Message": error_message}}

        exception = client("glue").exceptions.ClientError(err_response, "test")
        returned_exception = type(exception)

        glue_data_quality_mock_conn.exceptions.EntityNotFoundException = returned_exception
        glue_data_quality_mock_conn.update_data_quality_ruleset.side_effect = exception

        with pytest.raises(AirflowException, match=error_message):
            self.operator.execute({})

        glue_data_quality_mock_conn.update_data_quality_ruleset.assert_called_once_with(
            Description="update ruleset", Name=self.RULE_SET_NAME, Ruleset=self.RULE_SET
        )

    def test_validate_inputs(self):
        self.operator = GlueDataQualityOperator(
            task_id="create_data_quality_ruleset",
            name=self.RULE_SET_NAME,
            ruleset=self.RULE_SET,
        )

        assert self.operator.validate_inputs() is None

    def test_validate_inputs_error(self):
        self.operator = GlueDataQualityOperator(
            task_id="create_data_quality_ruleset",
            name=self.RULE_SET_NAME,
            ruleset='[ColumnLength "review_id" = 15]',
        )

        with pytest.raises(AttributeError, match="RuleSet must starts with Rules = \\[ and ends with \\]"):
            self.operator.validate_inputs()

    def test_template_fields(self):
        operator = GlueDataQualityOperator(
            task_id="create_data_quality_ruleset", name=self.RULE_SET_NAME, ruleset=self.RULE_SET
        )
        validate_template_fields(operator)

    def test_overwritten_conn_passed_to_hook(self):
        OVERWRITTEN_CONN = "new-conn-id"
        op = GlueDataQualityOperator(
            task_id="test_overwritten_conn_passed_to_hook",
            name=self.RULE_SET_NAME,
            ruleset=self.RULE_SET,
            aws_conn_id=OVERWRITTEN_CONN,
        )
        assert op.hook.aws_conn_id == OVERWRITTEN_CONN

    def test_default_conn_passed_to_hook(self):
        DEFAULT_CONN = "aws_default"
        op = GlueDataQualityOperator(
            task_id="test_default_conn_passed_to_hook", name=self.RULE_SET_NAME, ruleset=self.RULE_SET
        )
        assert op.hook.aws_conn_id == DEFAULT_CONN


class TestGlueDataQualityRuleSetEvaluationRunOperator:
    RUN_ID = "1234567890"
    DATA_SOURCE = {"GlueTable": {"DatabaseName": "TestDB", "TableName": "TestTable"}}
    ROLE = "role_arn"
    RULE_SET_NAMES = ["TestRuleSet"]

    @pytest.fixture
    def mock_conn(self) -> Generator[BaseAwsConnection, None, None]:
        with mock.patch.object(GlueDataQualityHook, "conn") as _conn:
            _conn.start_data_quality_ruleset_evaluation_run.return_value = {"RunId": self.RUN_ID}
            yield _conn

    @pytest.fixture
    def glue_data_quality_hook(self) -> Generator[GlueDataQualityHook, None, None]:
        with mock_aws():
            hook = GlueDataQualityHook(aws_conn_id="aws_default")
            yield hook

    def setup_method(self):
        self.operator = GlueDataQualityRuleSetEvaluationRunOperator(
            task_id="stat_evaluation_run",
            datasource=self.DATA_SOURCE,
            role=self.ROLE,
            rule_set_names=self.RULE_SET_NAMES,
            show_results=False,
        )
        self.operator.defer = mock.MagicMock()

    def test_init(self):
        assert self.operator.datasource == self.DATA_SOURCE
        assert self.operator.role == self.ROLE
        assert self.operator.rule_set_names == self.RULE_SET_NAMES

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_start_data_quality_ruleset_evaluation_run(self, glue_data_quality_mock_conn):
        glue_data_quality_mock_conn.get_data_quality_ruleset.return_value = {"Name": "TestRuleSet"}

        self.op = GlueDataQualityRuleSetEvaluationRunOperator(
            task_id="stat_evaluation_run",
            datasource=self.DATA_SOURCE,
            role=self.ROLE,
            number_of_workers=10,
            timeout=1000,
            rule_set_names=self.RULE_SET_NAMES,
            rule_set_evaluation_run_kwargs={"AdditionalRunOptions": {"CloudWatchMetricsEnabled": True}},
        )

        self.op.wait_for_completion = False
        self.op.execute({})

        glue_data_quality_mock_conn.start_data_quality_ruleset_evaluation_run.assert_called_once_with(
            DataSource=self.DATA_SOURCE,
            Role=self.ROLE,
            NumberOfWorkers=10,
            Timeout=1000,
            RulesetNames=self.RULE_SET_NAMES,
            AdditionalRunOptions={"CloudWatchMetricsEnabled": True},
        )

    def test_validate_inputs(self, mock_conn):
        mock_conn.get_data_quality_ruleset.return_value = {"Name": "TestRuleSet"}
        assert self.operator.validate_inputs() is None

    def test_validate_inputs_error(self, mock_conn):
        class RuleSetNotFoundException(Exception):
            pass

        mock_conn.exceptions.EntityNotFoundException = RuleSetNotFoundException
        mock_conn.get_data_quality_ruleset.side_effect = RuleSetNotFoundException()

        self.operator = GlueDataQualityRuleSetEvaluationRunOperator(
            task_id="stat_evaluation_run",
            datasource=self.DATA_SOURCE,
            role=self.ROLE,
            rule_set_names=["dummy"],
        )

        with pytest.raises(AirflowException, match="Following RulesetNames are not found \\['dummy'\\]"):
            self.operator.validate_inputs()

    @pytest.mark.parametrize(
        "wait_for_completion, deferrable",
        [
            pytest.param(False, False, id="no_wait"),
            pytest.param(True, False, id="wait"),
            pytest.param(False, True, id="defer"),
        ],
    )
    @mock.patch.object(GlueDataQualityHook, "get_waiter")
    def test_start_data_quality_ruleset_evaluation_run_wait_combinations(
        self, _, wait_for_completion, deferrable, mock_conn, glue_data_quality_hook
    ):
        mock_conn.get_data_quality_ruleset.return_value = {"Name": "TestRuleSet"}
        self.operator.wait_for_completion = wait_for_completion
        self.operator.deferrable = deferrable

        response = self.operator.execute({})

        assert response == self.RUN_ID
        assert glue_data_quality_hook.get_waiter.call_count == wait_for_completion
        assert self.operator.defer.call_count == deferrable

    def test_template_fields(self):
        validate_template_fields(self.operator)

    def test_overwritten_conn_passed_to_hook(self):
        OVERWRITTEN_CONN = "new-conn-id"
        op = GlueDataQualityRuleSetEvaluationRunOperator(
            task_id="test_overwritten_conn_passed_to_hook",
            datasource=self.DATA_SOURCE,
            role=self.ROLE,
            rule_set_names=self.RULE_SET_NAMES,
            show_results=False,
            aws_conn_id=OVERWRITTEN_CONN,
        )
        assert op.hook.aws_conn_id == OVERWRITTEN_CONN

    def test_default_conn_passed_to_hook(self):
        DEFAULT_CONN = "aws_default"
        op = GlueDataQualityRuleSetEvaluationRunOperator(
            task_id="test_default_conn_passed_to_hook",
            datasource=self.DATA_SOURCE,
            role=self.ROLE,
            rule_set_names=self.RULE_SET_NAMES,
            show_results=False,
        )
        assert op.hook.aws_conn_id == DEFAULT_CONN


class TestGlueDataQualityRuleRecommendationRunOperator:
    RUN_ID = "1234567890"
    DATA_SOURCE = {"GlueTable": {"DatabaseName": "TestDB", "TableName": "TestTable"}}
    ROLE = "role_arn"

    @pytest.fixture
    def mock_conn(self) -> Generator[BaseAwsConnection, None, None]:
        with mock.patch.object(GlueDataQualityHook, "conn") as _conn:
            _conn.start_data_quality_rule_recommendation_run.return_value = {"RunId": self.RUN_ID}
            yield _conn

    @pytest.fixture
    def glue_data_quality_hook(self) -> Generator[GlueDataQualityHook, None, None]:
        with mock_aws():
            hook = GlueDataQualityHook(aws_conn_id="aws_default")
            yield hook

    def setup_method(self):
        self.operator = GlueDataQualityRuleRecommendationRunOperator(
            task_id="start_recommendation_run",
            datasource=self.DATA_SOURCE,
            role=self.ROLE,
            show_results=False,
            recommendation_run_kwargs={"CreatedRulesetName": "test-ruleset"},
        )
        self.operator.defer = mock.MagicMock()

    def test_init(self):
        assert self.operator.datasource == self.DATA_SOURCE
        assert self.operator.role == self.ROLE
        assert self.operator.show_results is False
        assert self.operator.recommendation_run_kwargs == {"CreatedRulesetName": "test-ruleset"}

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_start_data_quality_rule_recommendation_run(self, glue_data_quality_mock_conn):
        self.op = GlueDataQualityRuleRecommendationRunOperator(
            task_id="start_recommendation_run",
            datasource=self.DATA_SOURCE,
            role=self.ROLE,
            number_of_workers=10,
            timeout=1000,
            recommendation_run_kwargs={"CreatedRulesetName": "test-ruleset"},
        )

        self.op.wait_for_completion = False
        self.op.execute({})

        glue_data_quality_mock_conn.start_data_quality_rule_recommendation_run.assert_called_once_with(
            DataSource=self.DATA_SOURCE,
            Role=self.ROLE,
            NumberOfWorkers=10,
            Timeout=1000,
            CreatedRulesetName="test-ruleset",
        )

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_start_data_quality_rule_recommendation_run_failed(self, glue_data_quality_mock_conn):
        created_ruleset_name = "test-ruleset"
        error_message = f"Ruleset {created_ruleset_name} already exists"

        err_response = {"Error": {"Code": "InvalidInputException", "Message": error_message}}

        exception = boto3.client("glue").exceptions.ClientError(
            err_response, "StartDataQualityRuleRecommendationRun"
        )
        returned_exception = type(exception)

        glue_data_quality_mock_conn.exceptions.InvalidInputException = returned_exception
        glue_data_quality_mock_conn.start_data_quality_rule_recommendation_run.side_effect = exception

        operator = GlueDataQualityRuleRecommendationRunOperator(
            task_id="stat_recommendation_run",
            datasource=self.DATA_SOURCE,
            role=self.ROLE,
            recommendation_run_kwargs={"CreatedRulesetName": created_ruleset_name},
        )
        operator.wait_for_completion = False

        with pytest.raises(
            AirflowException,
            match=f"AWS Glue data quality recommendation run failed: Ruleset {created_ruleset_name} already exists",
        ):
            operator.execute({})

    @pytest.mark.parametrize(
        "wait_for_completion, deferrable",
        [
            pytest.param(False, False, id="no_wait"),
            pytest.param(True, False, id="wait"),
            pytest.param(False, True, id="defer"),
        ],
    )
    @mock.patch.object(GlueDataQualityHook, "get_waiter")
    def test_start_data_quality_rule_recommendation_run_wait_combinations(
        self, _, wait_for_completion, deferrable, mock_conn, glue_data_quality_hook
    ):
        self.operator.wait_for_completion = wait_for_completion
        self.operator.deferrable = deferrable

        response = self.operator.execute({})

        assert response == self.RUN_ID
        assert glue_data_quality_hook.get_waiter.call_count == wait_for_completion
        assert self.operator.defer.call_count == deferrable

    def test_template_fields(self):
        validate_template_fields(self.operator)

    def test_overwritten_conn_passed_to_hook(self):
        OVERWRITTEN_CONN = "new-conn-id"
        op = GlueDataQualityRuleRecommendationRunOperator(
            task_id="test_overwritten_conn_passed_to_hook",
            datasource=self.DATA_SOURCE,
            role=self.ROLE,
            number_of_workers=10,
            timeout=1000,
            recommendation_run_kwargs={"CreatedRulesetName": "test-ruleset"},
            aws_conn_id=OVERWRITTEN_CONN,
        )
        assert op.hook.aws_conn_id == OVERWRITTEN_CONN

    def test_default_conn_passed_to_hook(self):
        DEFAULT_CONN = "aws_default"
        op = GlueDataQualityRuleRecommendationRunOperator(
            task_id="test_default_conn_passed_to_hook",
            datasource=self.DATA_SOURCE,
            role=self.ROLE,
            number_of_workers=10,
            timeout=1000,
            recommendation_run_kwargs={"CreatedRulesetName": "test-ruleset"},
        )
        assert op.hook.aws_conn_id == DEFAULT_CONN



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime
from unittest import mock
from unittest.mock import ANY

import pytest

from airflow.providers.amazon.aws.hooks.appflow import AppflowHook
from airflow.utils import timezone

FLOW_NAME = "flow0"
EXECUTION_ID = "ex_id"
CONNECTION_TYPE = "Salesforce"
REGION_NAME = "us-east-1"
AWS_CONN_ID = "aws_default"


@pytest.fixture
def hook():
    with mock.patch("airflow.providers.amazon.aws.hooks.appflow.AppflowHook.conn") as mock_conn:
        mock_conn.describe_flow.return_value = {
            "sourceFlowConfig": {"connectorType": CONNECTION_TYPE},
            "tasks": [],
            "triggerConfig": {"triggerProperties": None},
            "flowName": FLOW_NAME,
            "destinationFlowConfigList": {},
            "lastRunExecutionDetails": {
                "mostRecentExecutionStatus": "Successful",
                "mostRecentExecutionTime": datetime(3000, 1, 1, tzinfo=timezone.utc),
            },
        }
        mock_conn.update_flow.return_value = {}
        mock_conn.start_flow.return_value = {"executionId": EXECUTION_ID}
        mock_conn.describe_flow_execution_records.return_value = {
            "flowExecutions": [
                {
                    "executionId": EXECUTION_ID,
                    "executionResult": {"recordsProcessed": 1},
                    "executionStatus": "Successful",
                }
            ]
        }
        yield AppflowHook(aws_conn_id=AWS_CONN_ID, region_name=REGION_NAME)


def test_conn_attributes(hook):
    assert hasattr(hook, "conn")
    conn = hook.conn
    assert conn is hook.conn, "AppflowHook conn property non-cached"


def test_run_flow(hook):
    with mock.patch("airflow.providers.amazon.aws.waiters.base_waiter.BaseBotoWaiter.waiter"):
        hook.run_flow(flow_name=FLOW_NAME, poll_interval=0)
    hook.conn.describe_flow_execution_records.assert_called_with(flowName=FLOW_NAME)
    assert hook.conn.describe_flow_execution_records.call_count == 1
    hook.conn.start_flow.assert_called_once_with(flowName=FLOW_NAME)


def test_update_flow_filter(hook):
    tasks = [
        {
            "taskType": "Filter",
            "connectorOperator": {"Salesforce": "GREATER_THAN"},
            "sourceFields": ["col0"],
            "taskProperties": {"DATA_TYPE": "datetime", "VALUE": "1653523200000"},
        }
    ]
    hook.update_flow_filter(flow_name=FLOW_NAME, filter_tasks=tasks, set_trigger_ondemand=True)
    hook.conn.describe_flow.assert_called_with(flowName=FLOW_NAME)
    assert hook.conn.describe_flow.call_count == 1
    hook.conn.update_flow.assert_called_once_with(
        flowName=FLOW_NAME,
        tasks=tasks,
        description=ANY,
        destinationFlowConfigList=ANY,
        sourceFlowConfig=ANY,
        triggerConfig=ANY,
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.rds import (
    RdsCopyDbSnapshotOperator,
    RdsCreateDbInstanceOperator,
    RdsCreateDbSnapshotOperator,
    RdsDeleteDbInstanceOperator,
    RdsDeleteDbSnapshotOperator,
)
from airflow.providers.amazon.aws.sensors.rds import RdsSnapshotExistenceSensor
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

DAG_ID = "example_rds_snapshot"

sys_test_context_task = SystemTestContextBuilder().build()


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()

    rds_db_name = f"{test_context[ENV_ID_KEY]}_db"
    rds_instance_name = f"{test_context[ENV_ID_KEY]}-instance"
    rds_snapshot_name = f"{test_context[ENV_ID_KEY]}-snapshot"
    rds_snapshot_copy_name = f"{rds_snapshot_name}-copy"

    create_db_instance = RdsCreateDbInstanceOperator(
        task_id="create_db_instance",
        db_instance_identifier=rds_instance_name,
        db_instance_class="db.t4g.micro",
        engine="postgres",
        rds_kwargs={
            "MasterUsername": "rds_username",
            # NEVER store your production password in plaintext in a DAG like this.
            # Use Airflow Secrets or a secret manager for this in production.
            "MasterUserPassword": "rds_password",
            "AllocatedStorage": 20,
            "DBName": rds_db_name,
            "PubliclyAccessible": False,
        },
    )

    # [START howto_operator_rds_create_db_snapshot]
    create_snapshot = RdsCreateDbSnapshotOperator(
        task_id="create_snapshot",
        db_type="instance",
        db_identifier=rds_instance_name,
        db_snapshot_identifier=rds_snapshot_name,
    )
    # [END howto_operator_rds_create_db_snapshot]

    # [START howto_sensor_rds_snapshot_existence]
    snapshot_sensor = RdsSnapshotExistenceSensor(
        task_id="snapshot_sensor",
        db_type="instance",
        db_snapshot_identifier=rds_snapshot_name,
        target_statuses=["available"],
    )
    # [END howto_sensor_rds_snapshot_existence]

    # [START howto_operator_rds_copy_snapshot]
    copy_snapshot = RdsCopyDbSnapshotOperator(
        task_id="copy_snapshot",
        db_type="instance",
        source_db_snapshot_identifier=rds_snapshot_name,
        target_db_snapshot_identifier=rds_snapshot_copy_name,
    )
    # [END howto_operator_rds_copy_snapshot]

    # [START howto_operator_rds_delete_snapshot]
    delete_snapshot = RdsDeleteDbSnapshotOperator(
        task_id="delete_snapshot",
        db_type="instance",
        db_snapshot_identifier=rds_snapshot_name,
    )
    # [END howto_operator_rds_delete_snapshot]

    snapshot_copy_sensor = RdsSnapshotExistenceSensor(
        task_id="snapshot_copy_sensor",
        db_type="instance",
        db_snapshot_identifier=rds_snapshot_copy_name,
        target_statuses=["available"],
    )

    delete_snapshot_copy = RdsDeleteDbSnapshotOperator(
        task_id="delete_snapshot_copy",
        db_type="instance",
        db_snapshot_identifier=rds_snapshot_copy_name,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_db_instance = RdsDeleteDbInstanceOperator(
        task_id="delete_db_instance",
        db_instance_identifier=rds_instance_name,
        rds_kwargs={"SkipFinalSnapshot": True},
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        test_context,
        create_db_instance,
        # TEST BODY
        create_snapshot,
        snapshot_sensor,
        copy_snapshot,
        delete_snapshot,
        # TEST TEARDOWN
        snapshot_copy_sensor,
        delete_snapshot_copy,
        delete_db_instance,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow import DAG
from airflow.decorators import task, task_group
from airflow.models.baseoperator import chain
from airflow.providers.amazon.aws.hooks.glue import GlueDataQualityHook
from airflow.providers.amazon.aws.operators.athena import AthenaOperator
from airflow.providers.amazon.aws.operators.glue import (
    GlueDataQualityRuleRecommendationRunOperator,
    GlueDataQualityRuleSetEvaluationRunOperator,
)
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.sensors.glue import (
    GlueDataQualityRuleRecommendationRunSensor,
    GlueDataQualityRuleSetEvaluationRunSensor,
)
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import SystemTestContextBuilder

ROLE_ARN_KEY = "ROLE_ARN"
sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

DAG_ID = "example_glue_data_quality_with_recommendation"
SAMPLE_DATA = """"Alice",20
    "Bob",25
    "Charlie",30"""
SAMPLE_FILENAME = "airflow_sample.csv"


@task_group
def glue_data_quality_recommendation_workflow():
    # [START howto_operator_glue_data_quality_rule_recommendation_run]
    recommendation_run = GlueDataQualityRuleRecommendationRunOperator(
        task_id="recommendation_run",
        datasource={
            "GlueTable": {
                "TableName": athena_table,
                "DatabaseName": athena_database,
            }
        },
        role=test_context[ROLE_ARN_KEY],
        recommendation_run_kwargs={"CreatedRulesetName": rule_set_name},
    )
    # [END howto_operator_glue_data_quality_rule_recommendation_run]
    recommendation_run.wait_for_completion = False

    # [START howto_sensor_glue_data_quality_rule_recommendation_run]
    await_recommendation_run_sensor = GlueDataQualityRuleRecommendationRunSensor(
        task_id="await_recommendation_run_sensor",
        recommendation_run_id=recommendation_run.output,
    )
    # [END howto_sensor_glue_data_quality_rule_recommendation_run]

    start_evaluation_run = GlueDataQualityRuleSetEvaluationRunOperator(
        task_id="start_evaluation_run",
        datasource={
            "GlueTable": {
                "TableName": athena_table,
                "DatabaseName": athena_database,
            }
        },
        role=test_context[ROLE_ARN_KEY],
        rule_set_names=[rule_set_name],
    )
    start_evaluation_run.wait_for_completion = False

    await_evaluation_run_sensor = GlueDataQualityRuleSetEvaluationRunSensor(
        task_id="await_evaluation_run_sensor",
        evaluation_run_id=start_evaluation_run.output,
    )

    chain(
        recommendation_run, await_recommendation_run_sensor, start_evaluation_run, await_evaluation_run_sensor
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_ruleset(ruleset_name):
    hook = GlueDataQualityHook()
    hook.conn.delete_data_quality_ruleset(Name=ruleset_name)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]

    rule_set_name = f"{env_id}-recommendation-ruleset"
    s3_bucket = f"{env_id}-glue-dq-athena-bucket"
    athena_table = f"{env_id}_test_glue_dq_table"
    athena_database = f"{env_id}_glue_dq_default"

    query_create_database = f"CREATE DATABASE IF NOT EXISTS {athena_database}"
    query_create_table = f"""CREATE EXTERNAL TABLE IF NOT EXISTS {athena_database}.{athena_table}
            ( `name` string, `age` int )
            ROW FORMAT SERDE "org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe"
            WITH SERDEPROPERTIES ( "serialization.format" = ",", "field.delim" = "," )
            LOCATION "s3://{s3_bucket}//{athena_table}"
            TBLPROPERTIES ("has_encrypted_data"="false")
            """
    query_read_table = f"SELECT * from {athena_database}.{athena_table}"
    query_drop_table = f"DROP TABLE IF EXISTS {athena_database}.{athena_table}"
    query_drop_database = f"DROP DATABASE IF EXISTS {athena_database}"

    create_s3_bucket = S3CreateBucketOperator(task_id="create_s3_bucket", bucket_name=s3_bucket)

    upload_sample_data = S3CreateObjectOperator(
        task_id="upload_sample_data",
        s3_bucket=s3_bucket,
        s3_key=f"{athena_table}/{SAMPLE_FILENAME}",
        data=SAMPLE_DATA,
        replace=True,
    )

    create_database = AthenaOperator(
        task_id="create_database",
        query=query_create_database,
        database=athena_database,
        output_location=f"s3://{s3_bucket}/",
        sleep_time=1,
    )

    create_table = AthenaOperator(
        task_id="create_table",
        query=query_create_table,
        database=athena_database,
        output_location=f"s3://{s3_bucket}/",
        sleep_time=1,
    )

    drop_table = AthenaOperator(
        task_id="drop_table",
        query=query_drop_table,
        database=athena_database,
        output_location=f"s3://{s3_bucket}/",
        trigger_rule=TriggerRule.ALL_DONE,
        sleep_time=1,
    )

    drop_database = AthenaOperator(
        task_id="drop_database",
        query=query_drop_database,
        database=athena_database,
        output_location=f"s3://{s3_bucket}/",
        trigger_rule=TriggerRule.ALL_DONE,
        sleep_time=1,
    )

    delete_s3_bucket = S3DeleteBucketOperator(
        task_id="delete_s3_bucket",
        bucket_name=s3_bucket,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        test_context,
        create_s3_bucket,
        upload_sample_data,
        create_database,
        create_table,
        # TEST BODY
        glue_data_quality_recommendation_workflow(),
        # TEST TEARDOWN
        delete_ruleset(rule_set_name),
        drop_table,
        drop_database,
        delete_s3_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    #     This test needs watcher in order to properly mark success/failure
    #     when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.exceptions import AirflowException, AirflowNotFoundException
from airflow.providers.amazon.aws.hooks.rds import RdsHook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.amazon.aws.utils.rds import RdsDbType

if TYPE_CHECKING:
    from airflow.utils.context import Context


class RdsBaseSensor(AwsBaseSensor[RdsHook]):
    """Base operator that implements common functions for all sensors."""

    aws_hook_class = RdsHook
    ui_color = "#ddbb77"
    ui_fgcolor = "#ffffff"

    def __init__(self, *args, hook_params: dict | None = None, **kwargs):
        self.hook_params = hook_params or {}
        self.target_statuses: list[str] = []
        super().__init__(*args, **kwargs)


class RdsSnapshotExistenceSensor(RdsBaseSensor):
    """
    Waits for RDS snapshot with a specific status.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:RdsSnapshotExistenceSensor`

    :param db_type: Type of the DB - either "instance" or "cluster"
    :param db_snapshot_identifier: The identifier for the DB snapshot
    :param target_statuses: Target status of snapshot
    :param aws_conn_id: The Airflow connection used for AWS credentials.
         If this is ``None`` or empty then the default boto3 behaviour is used. If
         running Airflow in a distributed manner and aws_conn_id is None or
         empty, then default boto3 configuration would be used (and must be
         maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    template_fields: Sequence[str] = aws_template_fields(
        "db_snapshot_identifier",
        "target_statuses",
    )

    def __init__(
        self,
        *,
        db_type: str,
        db_snapshot_identifier: str,
        target_statuses: list[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.db_type = RdsDbType(db_type)
        self.db_snapshot_identifier = db_snapshot_identifier
        self.target_statuses = target_statuses or ["available"]

    def poke(self, context: Context):
        self.log.info(
            "Poking for statuses : %s\nfor snapshot %s", self.target_statuses, self.db_snapshot_identifier
        )
        try:
            if self.db_type.value == "instance":
                state = self.hook.get_db_snapshot_state(self.db_snapshot_identifier)
            else:
                state = self.hook.get_db_cluster_snapshot_state(self.db_snapshot_identifier)
        except AirflowNotFoundException:
            return False
        return state in self.target_statuses


class RdsExportTaskExistenceSensor(RdsBaseSensor):
    """
    Waits for RDS export task with a specific status.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:RdsExportTaskExistenceSensor`

    :param export_task_identifier: A unique identifier for the snapshot export task.
    :param target_statuses: Target status of export task
    :param error_statuses: Target error status of export task to fail the sensor
    """

    template_fields: Sequence[str] = aws_template_fields(
        "export_task_identifier", "target_statuses", "error_statuses"
    )

    def __init__(
        self,
        *,
        export_task_identifier: str,
        target_statuses: list[str] | None = None,
        error_statuses: list[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.export_task_identifier = export_task_identifier
        self.target_statuses = target_statuses or [
            "starting",
            "in_progress",
            "complete",
            "canceling",
            "canceled",
        ]
        self.error_statuses = error_statuses or ["failed"]

    def poke(self, context: Context):
        self.log.info(
            "Poking for statuses : %s\nfor export task %s", self.target_statuses, self.export_task_identifier
        )
        try:
            state = self.hook.get_export_task_state(self.export_task_identifier)
            if state in self.error_statuses:
                raise AirflowException(
                    f"Export task {self.export_task_identifier} failed with status {state}"
                )

        except AirflowNotFoundException:
            return False
        return state in self.target_statuses


class RdsDbSensor(RdsBaseSensor):
    """
    Waits for an RDS instance or cluster to enter one of a number of states.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:RdsDbSensor`

    :param db_type: Type of the DB - either "instance" or "cluster" (default: 'instance')
    :param db_identifier: The AWS identifier for the DB
    :param target_statuses: Target status of DB
    """

    template_fields: Sequence[str] = aws_template_fields(
        "db_identifier",
        "db_type",
        "target_statuses",
    )

    def __init__(
        self,
        *,
        db_identifier: str,
        db_type: RdsDbType | str = RdsDbType.INSTANCE,
        target_statuses: list[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.db_identifier = db_identifier
        self.target_statuses = target_statuses or ["available"]
        self.db_type = db_type

    def poke(self, context: Context):
        db_type = RdsDbType(self.db_type)
        self.log.info(
            "Poking for statuses : %s\nfor db instance %s", self.target_statuses, self.db_identifier
        )
        try:
            if db_type == RdsDbType.INSTANCE:
                state = self.hook.get_db_instance_state(self.db_identifier)
            else:
                state = self.hook.get_db_cluster_state(self.db_identifier)
        except AirflowNotFoundException:
            return False
        return state in self.target_statuses


__all__ = [
    "RdsExportTaskExistenceSensor",
    "RdsDbSensor",
    "RdsSnapshotExistenceSensor",
]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from unittest import mock
from unittest.mock import patch

import pytest
from moto import mock_aws

from airflow.exceptions import TaskDeferred
from airflow.models import DAG
from airflow.providers.amazon.aws.hooks.rds import RdsHook
from airflow.providers.amazon.aws.operators.rds import (
    RdsBaseOperator,
    RdsCancelExportTaskOperator,
    RdsCopyDbSnapshotOperator,
    RdsCreateDbInstanceOperator,
    RdsCreateDbSnapshotOperator,
    RdsCreateEventSubscriptionOperator,
    RdsDeleteDbInstanceOperator,
    RdsDeleteDbSnapshotOperator,
    RdsDeleteEventSubscriptionOperator,
    RdsStartDbOperator,
    RdsStartExportTaskOperator,
    RdsStopDbOperator,
)
from airflow.providers.amazon.aws.triggers.rds import RdsDbAvailableTrigger, RdsDbStoppedTrigger
from airflow.utils import timezone

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.hooks.base_aws import AwsGenericHook

DEFAULT_DATE = timezone.datetime(2019, 1, 1)

AWS_CONN = "amazon_default"

REGION = "us-east-1"

DB_INSTANCE_NAME = "my-db-instance"
DB_CLUSTER_NAME = "my-db-cluster"

DB_INSTANCE_SNAPSHOT = "my-db-instance-snap"
DB_CLUSTER_SNAPSHOT = "my-db-cluster-snap"

DB_INSTANCE_SNAPSHOT_COPY = "my-db-instance-snap-copy"
DB_CLUSTER_SNAPSHOT_COPY = "my-db-cluster-snap-copy"

EXPORT_TASK_NAME = "my-db-instance-snap-export"
EXPORT_TASK_SOURCE = "arn:aws:rds:es-east-1::snapshot:my-db-instance-snap"
EXPORT_TASK_ROLE_NAME = "MyRole"
EXPORT_TASK_ROLE_ARN = "arn:aws:iam:es-east-1::role/MyRole"
EXPORT_TASK_KMS = "arn:aws:kms:es-east-1::key/*****-****-****-****-********"
EXPORT_TASK_BUCKET = "my-exports-bucket"

SUBSCRIPTION_NAME = "my-db-instance-subscription"
SUBSCRIPTION_TOPIC = "arn:aws:sns:us-east-1::MyTopic"


def _create_db_instance(hook: RdsHook):
    hook.conn.create_db_instance(
        DBInstanceIdentifier=DB_INSTANCE_NAME,
        DBInstanceClass="db.m4.large",
        Engine="postgres",
    )
    if not hook.conn.describe_db_instances()["DBInstances"]:
        raise ValueError("AWS not properly mocked")


def _create_db_cluster(hook: RdsHook):
    hook.conn.create_db_cluster(
        DBClusterIdentifier=DB_CLUSTER_NAME,
        Engine="mysql",
        MasterUsername="admin",
        MasterUserPassword="admin-pass",
    )
    if not hook.conn.describe_db_clusters()["DBClusters"]:
        raise ValueError("AWS not properly mocked")


def _create_db_instance_snapshot(hook: RdsHook):
    hook.conn.create_db_snapshot(
        DBInstanceIdentifier=DB_INSTANCE_NAME,
        DBSnapshotIdentifier=DB_INSTANCE_SNAPSHOT,
    )
    if not hook.conn.describe_db_snapshots()["DBSnapshots"]:
        raise ValueError("AWS not properly mocked")


def _create_db_cluster_snapshot(hook: RdsHook):
    hook.conn.create_db_cluster_snapshot(
        DBClusterIdentifier=DB_CLUSTER_NAME,
        DBClusterSnapshotIdentifier=DB_CLUSTER_SNAPSHOT,
    )
    if not hook.conn.describe_db_cluster_snapshots()["DBClusterSnapshots"]:
        raise ValueError("AWS not properly mocked")


def _start_export_task(hook: RdsHook):
    hook.conn.start_export_task(
        ExportTaskIdentifier=EXPORT_TASK_NAME,
        SourceArn=EXPORT_TASK_SOURCE,
        IamRoleArn=EXPORT_TASK_ROLE_ARN,
        KmsKeyId=EXPORT_TASK_KMS,
        S3BucketName=EXPORT_TASK_BUCKET,
    )
    if not hook.conn.describe_export_tasks()["ExportTasks"]:
        raise ValueError("AWS not properly mocked")


def _create_event_subscription(hook: RdsHook):
    hook.conn.create_event_subscription(
        SubscriptionName=SUBSCRIPTION_NAME,
        SnsTopicArn=SUBSCRIPTION_TOPIC,
        SourceType="db-instance",
        SourceIds=[DB_INSTANCE_NAME],
    )
    if not hook.conn.describe_event_subscriptions()["EventSubscriptionsList"]:
        raise ValueError("AWS not properly mocked")


def _patch_hook_get_connection(hook: AwsGenericHook) -> None:
    # We're mocking all actual AWS calls and don't need a connection. This
    # avoids an Airflow warning about connection cannot be found.
    hook.get_connection = lambda _: None  # type: ignore[assignment,return-value]


def _create_kms_key() -> str:
    import boto3

    conn = boto3.client("kms", region_name="us-east-1")
    metadata = conn.create_key(Policy="my policy")["KeyMetadata"]
    return metadata["Arn"]


class TestBaseRdsOperator:
    dag = None
    op = None

    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.op = RdsBaseOperator(task_id="test_task", aws_conn_id="aws_default", dag=cls.dag)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.op

    def test_hook_attribute(self):
        assert hasattr(self.op, "hook")
        assert self.op.hook.__class__.__name__ == "RdsHook"

    def test_overwritten_conn_passed_to_hook(self):
        OVERWRITTEN_CONN = "new-conn-id"
        op = RdsBaseOperator(
            task_id="test_overwritten_conn_passed_to_hook_task", aws_conn_id=OVERWRITTEN_CONN, dag=self.dag
        )
        assert op.hook.aws_conn_id == OVERWRITTEN_CONN

    def test_default_conn_passed_to_hook(self):
        DEFAULT_CONN = "aws_default"
        op = RdsBaseOperator(task_id="test_default_conn_passed_to_hook_task", dag=self.dag)
        assert op.hook.aws_conn_id == DEFAULT_CONN


class TestRdsCreateDbSnapshotOperator:
    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.hook = RdsHook(aws_conn_id=AWS_CONN, region_name="us-east-1")
        _patch_hook_get_connection(cls.hook)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.hook

    @mock_aws
    def test_create_db_instance_snapshot(self):
        _create_db_instance(self.hook)
        instance_snapshot_operator = RdsCreateDbSnapshotOperator(
            task_id="test_instance",
            db_type="instance",
            db_snapshot_identifier=DB_INSTANCE_SNAPSHOT,
            db_identifier=DB_INSTANCE_NAME,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
        )
        _patch_hook_get_connection(instance_snapshot_operator.hook)
        instance_snapshot_operator.execute(None)

        result = self.hook.conn.describe_db_snapshots(DBSnapshotIdentifier=DB_INSTANCE_SNAPSHOT)
        instance_snapshots = result.get("DBSnapshots")

        assert instance_snapshots
        assert len(instance_snapshots) == 1

    @mock_aws
    @patch.object(RdsHook, "wait_for_db_snapshot_state")
    def test_create_db_instance_snapshot_no_wait(self, mock_wait):
        _create_db_instance(self.hook)
        instance_snapshot_operator = RdsCreateDbSnapshotOperator(
            task_id="test_instance_no_wait",
            db_type="instance",
            db_snapshot_identifier=DB_INSTANCE_SNAPSHOT,
            db_identifier=DB_INSTANCE_NAME,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            wait_for_completion=False,
        )
        _patch_hook_get_connection(instance_snapshot_operator.hook)
        instance_snapshot_operator.execute(None)

        result = self.hook.conn.describe_db_snapshots(DBSnapshotIdentifier=DB_INSTANCE_SNAPSHOT)
        instance_snapshots = result.get("DBSnapshots")

        assert instance_snapshots
        assert len(instance_snapshots) == 1
        mock_wait.assert_not_called()

    @mock_aws
    def test_create_db_cluster_snapshot(self):
        _create_db_cluster(self.hook)
        cluster_snapshot_operator = RdsCreateDbSnapshotOperator(
            task_id="test_cluster",
            db_type="cluster",
            db_snapshot_identifier=DB_CLUSTER_SNAPSHOT,
            db_identifier=DB_CLUSTER_NAME,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
        )
        _patch_hook_get_connection(cluster_snapshot_operator.hook)
        cluster_snapshot_operator.execute(None)

        result = self.hook.conn.describe_db_cluster_snapshots(DBClusterSnapshotIdentifier=DB_CLUSTER_SNAPSHOT)
        cluster_snapshots = result.get("DBClusterSnapshots")

        assert cluster_snapshots
        assert len(cluster_snapshots) == 1

    @mock_aws
    @patch.object(RdsHook, "wait_for_db_cluster_snapshot_state")
    def test_create_db_cluster_snapshot_no_wait(self, mock_wait):
        _create_db_cluster(self.hook)
        cluster_snapshot_operator = RdsCreateDbSnapshotOperator(
            task_id="test_cluster_no_wait",
            db_type="cluster",
            db_snapshot_identifier=DB_CLUSTER_SNAPSHOT,
            db_identifier=DB_CLUSTER_NAME,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            wait_for_completion=False,
        )
        _patch_hook_get_connection(cluster_snapshot_operator.hook)
        cluster_snapshot_operator.execute(None)

        result = self.hook.conn.describe_db_cluster_snapshots(DBClusterSnapshotIdentifier=DB_CLUSTER_SNAPSHOT)
        cluster_snapshots = result.get("DBClusterSnapshots")

        assert cluster_snapshots
        assert len(cluster_snapshots) == 1
        mock_wait.assert_not_called()

    def test_template_fields(self):
        operator = RdsCreateDbSnapshotOperator(
            task_id="test_instance_",
            db_type="instance",
            db_snapshot_identifier=DB_INSTANCE_SNAPSHOT,
            db_identifier=DB_INSTANCE_NAME,
            aws_conn_id=AWS_CONN,
            region_name=REGION,
        )
        validate_template_fields(operator)


class TestRdsCopyDbSnapshotOperator:
    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.hook = RdsHook(aws_conn_id=AWS_CONN, region_name="us-east-1")
        _patch_hook_get_connection(cls.hook)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.hook

    @mock_aws
    def test_copy_db_instance_snapshot(self):
        kms_key_arn = _create_kms_key()
        _create_db_instance(self.hook)
        _create_db_instance_snapshot(self.hook)

        instance_snapshot_operator = RdsCopyDbSnapshotOperator(
            task_id="test_instance",
            db_type="instance",
            source_db_snapshot_identifier=DB_INSTANCE_SNAPSHOT,
            target_db_snapshot_identifier=DB_INSTANCE_SNAPSHOT_COPY,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            kms_key_id=kms_key_arn,
        )
        _patch_hook_get_connection(instance_snapshot_operator.hook)
        instance_snapshot_operator.execute(None)
        result = self.hook.conn.describe_db_snapshots(DBSnapshotIdentifier=DB_INSTANCE_SNAPSHOT_COPY)
        instance_snapshots = result.get("DBSnapshots")

        assert instance_snapshots
        assert len(instance_snapshots) == 1

    @mock_aws
    @patch.object(RdsHook, "wait_for_db_snapshot_state")
    def test_copy_db_instance_snapshot_no_wait(self, mock_await_status):
        kms_key_arn = _create_kms_key()
        _create_db_instance(self.hook)
        _create_db_instance_snapshot(self.hook)

        instance_snapshot_operator = RdsCopyDbSnapshotOperator(
            task_id="test_instance_no_wait",
            db_type="instance",
            source_db_snapshot_identifier=DB_INSTANCE_SNAPSHOT,
            target_db_snapshot_identifier=DB_INSTANCE_SNAPSHOT_COPY,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            wait_for_completion=False,
            kms_key_id=kms_key_arn,
        )
        _patch_hook_get_connection(instance_snapshot_operator.hook)
        instance_snapshot_operator.execute(None)
        result = self.hook.conn.describe_db_snapshots(DBSnapshotIdentifier=DB_INSTANCE_SNAPSHOT_COPY)
        instance_snapshots = result.get("DBSnapshots")

        assert instance_snapshots
        assert len(instance_snapshots) == 1
        mock_await_status.assert_not_called()

    @mock_aws
    def test_copy_db_cluster_snapshot(self):
        kms_key_arn = _create_kms_key()
        _create_db_cluster(self.hook)
        _create_db_cluster_snapshot(self.hook)

        cluster_snapshot_operator = RdsCopyDbSnapshotOperator(
            task_id="test_cluster",
            db_type="cluster",
            source_db_snapshot_identifier=DB_CLUSTER_SNAPSHOT,
            target_db_snapshot_identifier=DB_CLUSTER_SNAPSHOT_COPY,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            kms_key_id=kms_key_arn,
        )
        _patch_hook_get_connection(cluster_snapshot_operator.hook)
        cluster_snapshot_operator.execute(None)
        result = self.hook.conn.describe_db_cluster_snapshots(
            DBClusterSnapshotIdentifier=DB_CLUSTER_SNAPSHOT_COPY
        )
        cluster_snapshots = result.get("DBClusterSnapshots")

        assert cluster_snapshots
        assert len(cluster_snapshots) == 1

    @mock_aws
    @patch.object(RdsHook, "wait_for_db_snapshot_state")
    def test_copy_db_cluster_snapshot_no_wait(self, mock_await_status):
        kms_key_arn = _create_kms_key()
        _create_db_cluster(self.hook)
        _create_db_cluster_snapshot(self.hook)

        cluster_snapshot_operator = RdsCopyDbSnapshotOperator(
            task_id="test_cluster_no_wait",
            db_type="cluster",
            source_db_snapshot_identifier=DB_CLUSTER_SNAPSHOT,
            target_db_snapshot_identifier=DB_CLUSTER_SNAPSHOT_COPY,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            kms_key_id=kms_key_arn,
        )
        _patch_hook_get_connection(cluster_snapshot_operator.hook)
        cluster_snapshot_operator.execute(None)
        result = self.hook.conn.describe_db_cluster_snapshots(
            DBClusterSnapshotIdentifier=DB_CLUSTER_SNAPSHOT_COPY
        )
        cluster_snapshots = result.get("DBClusterSnapshots")

        assert cluster_snapshots
        assert len(cluster_snapshots) == 1
        mock_await_status.assert_not_called()

    def test_template_fields(self):
        operator = RdsCopyDbSnapshotOperator(
            task_id="test_cluster_no_wait",
            db_type="cluster",
            source_db_snapshot_identifier=DB_CLUSTER_SNAPSHOT,
            target_db_snapshot_identifier=DB_CLUSTER_SNAPSHOT_COPY,
            aws_conn_id=AWS_CONN,
            region_name=REGION,
        )
        validate_template_fields(operator)


class TestRdsDeleteDbSnapshotOperator:
    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.hook = RdsHook(aws_conn_id=AWS_CONN, region_name="us-east-1")
        _patch_hook_get_connection(cls.hook)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.hook

    @mock_aws
    def test_delete_db_instance_snapshot(self):
        _create_db_instance(self.hook)
        _create_db_instance_snapshot(self.hook)

        instance_snapshot_operator = RdsDeleteDbSnapshotOperator(
            task_id="test_instance",
            db_type="instance",
            db_snapshot_identifier=DB_INSTANCE_SNAPSHOT,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
        )
        _patch_hook_get_connection(instance_snapshot_operator.hook)
        with patch.object(instance_snapshot_operator.hook, "wait_for_db_snapshot_state") as mock_wait:
            instance_snapshot_operator.execute(None)
        mock_wait.assert_called_once_with(DB_INSTANCE_SNAPSHOT, target_state="deleted")

        with pytest.raises(self.hook.conn.exceptions.ClientError):
            self.hook.conn.describe_db_snapshots(DBSnapshotIdentifier=DB_INSTANCE_SNAPSHOT)

    @mock_aws
    def test_delete_db_instance_snapshot_no_wait(self):
        """
        Check that the operator does not wait for the DB instance snapshot delete operation to complete when
        wait_for_completion=False
        """
        _create_db_instance(self.hook)
        _create_db_instance_snapshot(self.hook)

        instance_snapshot_operator = RdsDeleteDbSnapshotOperator(
            task_id="test_delete_db_instance_snapshot_no_wait",
            db_type="instance",
            db_snapshot_identifier=DB_INSTANCE_SNAPSHOT,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            wait_for_completion=False,
        )
        _patch_hook_get_connection(instance_snapshot_operator.hook)
        with patch.object(instance_snapshot_operator.hook, "wait_for_db_snapshot_state") as mock_wait:
            instance_snapshot_operator.execute(None)
        mock_wait.assert_not_called()

        with pytest.raises(self.hook.conn.exceptions.ClientError):
            self.hook.conn.describe_db_snapshots(DBSnapshotIdentifier=DB_INSTANCE_SNAPSHOT)

    @mock_aws
    def test_delete_db_cluster_snapshot(self):
        _create_db_cluster(self.hook)
        _create_db_cluster_snapshot(self.hook)

        cluster_snapshot_operator = RdsDeleteDbSnapshotOperator(
            task_id="test_cluster",
            db_type="cluster",
            db_snapshot_identifier=DB_CLUSTER_SNAPSHOT,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
        )
        _patch_hook_get_connection(cluster_snapshot_operator.hook)
        with patch.object(cluster_snapshot_operator.hook, "wait_for_db_cluster_snapshot_state") as mock_wait:
            cluster_snapshot_operator.execute(None)
        mock_wait.assert_called_once_with(DB_CLUSTER_SNAPSHOT, target_state="deleted")

        with pytest.raises(self.hook.conn.exceptions.ClientError):
            self.hook.conn.describe_db_cluster_snapshots(DBClusterSnapshotIdentifier=DB_CLUSTER_SNAPSHOT)

    @mock_aws
    def test_delete_db_cluster_snapshot_no_wait(self):
        """
        Check that the operator does not wait for the DB cluster snapshot delete operation to complete when
        wait_for_completion=False
        """
        _create_db_cluster(self.hook)
        _create_db_cluster_snapshot(self.hook)

        cluster_snapshot_operator = RdsDeleteDbSnapshotOperator(
            task_id="test_delete_db_cluster_snapshot_no_wait",
            db_type="cluster",
            db_snapshot_identifier=DB_CLUSTER_SNAPSHOT,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            wait_for_completion=False,
        )
        _patch_hook_get_connection(cluster_snapshot_operator.hook)
        with patch.object(cluster_snapshot_operator.hook, "wait_for_db_cluster_snapshot_state") as mock_wait:
            cluster_snapshot_operator.execute(None)
        mock_wait.assert_not_called()

        with pytest.raises(self.hook.conn.exceptions.ClientError):
            self.hook.conn.describe_db_cluster_snapshots(DBClusterSnapshotIdentifier=DB_CLUSTER_SNAPSHOT)

    def test_template_fields(self):
        operator = RdsDeleteDbSnapshotOperator(
            task_id="test_delete_db_cluster_snapshot_no_wait",
            db_type="cluster",
            db_snapshot_identifier=DB_CLUSTER_SNAPSHOT,
            aws_conn_id=AWS_CONN,
            wait_for_completion=False,
            region_name=REGION,
        )
        validate_template_fields(operator)


class TestRdsStartExportTaskOperator:
    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.hook = RdsHook(aws_conn_id=AWS_CONN, region_name="us-east-1")
        _patch_hook_get_connection(cls.hook)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.hook

    @mock_aws
    def test_start_export_task(self):
        _create_db_instance(self.hook)
        _create_db_instance_snapshot(self.hook)

        start_export_operator = RdsStartExportTaskOperator(
            task_id="test_start",
            export_task_identifier=EXPORT_TASK_NAME,
            source_arn=EXPORT_TASK_SOURCE,
            iam_role_arn=EXPORT_TASK_ROLE_ARN,
            kms_key_id=EXPORT_TASK_KMS,
            s3_bucket_name=EXPORT_TASK_BUCKET,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
        )
        _patch_hook_get_connection(start_export_operator.hook)
        start_export_operator.execute(None)

        result = self.hook.conn.describe_export_tasks(ExportTaskIdentifier=EXPORT_TASK_NAME)
        export_tasks = result.get("ExportTasks")

        assert export_tasks
        assert len(export_tasks) == 1
        assert export_tasks[0]["Status"] == "complete"

    @mock_aws
    @patch.object(RdsHook, "wait_for_export_task_state")
    def test_start_export_task_no_wait(self, mock_await_status):
        _create_db_instance(self.hook)
        _create_db_instance_snapshot(self.hook)

        start_export_operator = RdsStartExportTaskOperator(
            task_id="test_start_no_wait",
            export_task_identifier=EXPORT_TASK_NAME,
            source_arn=EXPORT_TASK_SOURCE,
            iam_role_arn=EXPORT_TASK_ROLE_ARN,
            kms_key_id=EXPORT_TASK_KMS,
            s3_bucket_name=EXPORT_TASK_BUCKET,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            wait_for_completion=False,
        )
        _patch_hook_get_connection(start_export_operator.hook)
        start_export_operator.execute(None)

        result = self.hook.conn.describe_export_tasks(ExportTaskIdentifier=EXPORT_TASK_NAME)
        export_tasks = result.get("ExportTasks")

        assert export_tasks
        assert len(export_tasks) == 1
        assert export_tasks[0]["Status"] == "complete"
        mock_await_status.assert_not_called()

    def test_template_fields(self):
        operator = RdsStartExportTaskOperator(
            task_id="test_start_no_wait",
            export_task_identifier=EXPORT_TASK_NAME,
            source_arn=EXPORT_TASK_SOURCE,
            iam_role_arn=EXPORT_TASK_ROLE_ARN,
            kms_key_id=EXPORT_TASK_KMS,
            s3_bucket_name=EXPORT_TASK_BUCKET,
            aws_conn_id=AWS_CONN,
            wait_for_completion=False,
            region_name=REGION,
        )
        validate_template_fields(operator)


class TestRdsCancelExportTaskOperator:
    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.hook = RdsHook(aws_conn_id=AWS_CONN, region_name="us-east-1")
        _patch_hook_get_connection(cls.hook)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.hook

    @mock_aws
    def test_cancel_export_task(self):
        _create_db_instance(self.hook)
        _create_db_instance_snapshot(self.hook)
        _start_export_task(self.hook)

        cancel_export_operator = RdsCancelExportTaskOperator(
            task_id="test_cancel",
            export_task_identifier=EXPORT_TASK_NAME,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
        )
        _patch_hook_get_connection(cancel_export_operator.hook)
        cancel_export_operator.execute(None)

        result = self.hook.conn.describe_export_tasks(ExportTaskIdentifier=EXPORT_TASK_NAME)
        export_tasks = result.get("ExportTasks")

        assert export_tasks
        assert len(export_tasks) == 1
        assert export_tasks[0]["Status"] == "canceled"

    @mock_aws
    @patch.object(RdsHook, "wait_for_export_task_state")
    def test_cancel_export_task_no_wait(self, mock_await_status):
        _create_db_instance(self.hook)
        _create_db_instance_snapshot(self.hook)
        _start_export_task(self.hook)

        cancel_export_operator = RdsCancelExportTaskOperator(
            task_id="test_cancel_no_wait",
            export_task_identifier=EXPORT_TASK_NAME,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            wait_for_completion=False,
        )
        _patch_hook_get_connection(cancel_export_operator.hook)
        cancel_export_operator.execute(None)

        result = self.hook.conn.describe_export_tasks(ExportTaskIdentifier=EXPORT_TASK_NAME)
        export_tasks = result.get("ExportTasks")

        assert export_tasks
        assert len(export_tasks) == 1
        assert export_tasks[0]["Status"] == "canceled"
        mock_await_status.assert_not_called()

    def test_template_fields(self):
        operator = RdsCancelExportTaskOperator(
            task_id="test_cancel",
            export_task_identifier=EXPORT_TASK_NAME,
            aws_conn_id=AWS_CONN,
            region_name=REGION,
        )
        validate_template_fields(operator)


class TestRdsCreateEventSubscriptionOperator:
    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.hook = RdsHook(aws_conn_id=AWS_CONN, region_name="us-east-1")
        _patch_hook_get_connection(cls.hook)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.hook

    @mock_aws
    def test_create_event_subscription(self):
        _create_db_instance(self.hook)

        create_subscription_operator = RdsCreateEventSubscriptionOperator(
            task_id="test_create",
            subscription_name=SUBSCRIPTION_NAME,
            sns_topic_arn=SUBSCRIPTION_TOPIC,
            source_type="db-instance",
            source_ids=[DB_INSTANCE_NAME],
            aws_conn_id=AWS_CONN,
            dag=self.dag,
        )
        _patch_hook_get_connection(create_subscription_operator.hook)
        create_subscription_operator.execute(None)

        result = self.hook.conn.describe_event_subscriptions(SubscriptionName=SUBSCRIPTION_NAME)
        subscriptions = result.get("EventSubscriptionsList")

        assert subscriptions
        assert len(subscriptions) == 1
        assert subscriptions[0]["Status"] == "active"

    @mock_aws
    @patch.object(RdsHook, "wait_for_event_subscription_state")
    def test_create_event_subscription_no_wait(self, mock_await_status):
        _create_db_instance(self.hook)

        create_subscription_operator = RdsCreateEventSubscriptionOperator(
            task_id="test_create_no_wait",
            subscription_name=SUBSCRIPTION_NAME,
            sns_topic_arn=SUBSCRIPTION_TOPIC,
            source_type="db-instance",
            source_ids=[DB_INSTANCE_NAME],
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            wait_for_completion=False,
        )
        _patch_hook_get_connection(create_subscription_operator.hook)
        create_subscription_operator.execute(None)

        result = self.hook.conn.describe_event_subscriptions(SubscriptionName=SUBSCRIPTION_NAME)
        subscriptions = result.get("EventSubscriptionsList")

        assert subscriptions
        assert len(subscriptions) == 1
        assert subscriptions[0]["Status"] == "active"
        mock_await_status.assert_not_called()

    def test_template_fields(self):
        operator = RdsCreateEventSubscriptionOperator(
            task_id="test_create",
            subscription_name=SUBSCRIPTION_NAME,
            sns_topic_arn=SUBSCRIPTION_TOPIC,
            source_type="db-instance",
            source_ids=[DB_INSTANCE_NAME],
            aws_conn_id=AWS_CONN,
            region_name=REGION,
        )
        validate_template_fields(operator)


class TestRdsDeleteEventSubscriptionOperator:
    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.hook = RdsHook(aws_conn_id=AWS_CONN, region_name="us-east-1")
        _patch_hook_get_connection(cls.hook)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.hook

    @mock_aws
    def test_delete_event_subscription(self):
        _create_event_subscription(self.hook)

        delete_subscription_operator = RdsDeleteEventSubscriptionOperator(
            task_id="test_delete",
            subscription_name=SUBSCRIPTION_NAME,
            aws_conn_id=AWS_CONN,
            dag=self.dag,
        )
        _patch_hook_get_connection(delete_subscription_operator.hook)
        delete_subscription_operator.execute(None)

        with pytest.raises(self.hook.conn.exceptions.ClientError):
            self.hook.conn.describe_event_subscriptions(SubscriptionName=EXPORT_TASK_NAME)

    def test_template_fields(self):
        operator = RdsDeleteEventSubscriptionOperator(
            task_id="test_delete",
            subscription_name=SUBSCRIPTION_NAME,
            aws_conn_id=AWS_CONN,
            region_name=REGION,
        )
        validate_template_fields(operator)


class TestRdsCreateDbInstanceOperator:
    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.hook = RdsHook(aws_conn_id=AWS_CONN, region_name="us-east-1")
        _patch_hook_get_connection(cls.hook)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.hook

    @mock_aws
    def test_create_db_instance(self):
        create_db_instance_operator = RdsCreateDbInstanceOperator(
            task_id="test_create_db_instance",
            db_instance_identifier=DB_INSTANCE_NAME,
            db_instance_class="db.m5.large",
            engine="postgres",
            rds_kwargs={
                "DBName": DB_INSTANCE_NAME,
            },
            aws_conn_id=AWS_CONN,
            dag=self.dag,
        )
        _patch_hook_get_connection(create_db_instance_operator.hook)
        create_db_instance_operator.execute(None)

        result = self.hook.conn.describe_db_instances(DBInstanceIdentifier=DB_INSTANCE_NAME)
        db_instances = result.get("DBInstances")

        assert db_instances
        assert len(db_instances) == 1
        assert db_instances[0]["DBInstanceStatus"] == "available"

    @mock_aws
    @patch.object(RdsHook, "wait_for_db_instance_state")
    def test_create_db_instance_no_wait(self, mock_await_status):
        create_db_instance_operator = RdsCreateDbInstanceOperator(
            task_id="test_create_db_instance_no_wait",
            db_instance_identifier=DB_INSTANCE_NAME,
            db_instance_class="db.m5.large",
            engine="postgres",
            rds_kwargs={
                "DBName": DB_INSTANCE_NAME,
            },
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            wait_for_completion=False,
        )
        _patch_hook_get_connection(create_db_instance_operator.hook)
        create_db_instance_operator.execute(None)

        result = self.hook.conn.describe_db_instances(DBInstanceIdentifier=DB_INSTANCE_NAME)
        db_instances = result.get("DBInstances")

        assert db_instances
        assert len(db_instances) == 1
        assert db_instances[0]["DBInstanceStatus"] == "available"
        mock_await_status.assert_not_called()

    def test_template_fields(self):
        operator = RdsCreateDbInstanceOperator(
            task_id="test_create_db_instance",
            db_instance_identifier=DB_INSTANCE_NAME,
            db_instance_class="db.m5.large",
            engine="postgres",
            rds_kwargs={
                "DBName": DB_INSTANCE_NAME,
            },
            aws_conn_id=AWS_CONN,
            region_name=REGION,
        )
        validate_template_fields(operator)


class TestRdsDeleteDbInstanceOperator:
    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.hook = RdsHook(aws_conn_id=AWS_CONN, region_name="us-east-1")
        _patch_hook_get_connection(cls.hook)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.hook

    @mock_aws
    def test_delete_db_instance(self):
        _create_db_instance(self.hook)

        delete_db_instance_operator = RdsDeleteDbInstanceOperator(
            task_id="test_delete_db_instance",
            db_instance_identifier=DB_INSTANCE_NAME,
            rds_kwargs={
                "SkipFinalSnapshot": True,
            },
            aws_conn_id=AWS_CONN,
            dag=self.dag,
        )
        _patch_hook_get_connection(delete_db_instance_operator.hook)
        delete_db_instance_operator.execute(None)

        with pytest.raises(self.hook.conn.exceptions.ClientError):
            self.hook.conn.describe_db_instances(DBInstanceIdentifier=DB_INSTANCE_NAME)

    @mock_aws
    @patch.object(RdsHook, "wait_for_db_instance_state")
    def test_delete_db_instance_no_wait(self, mock_await_status):
        _create_db_instance(self.hook)

        delete_db_instance_operator = RdsDeleteDbInstanceOperator(
            task_id="test_delete_db_instance_no_wait",
            db_instance_identifier=DB_INSTANCE_NAME,
            rds_kwargs={
                "SkipFinalSnapshot": True,
            },
            aws_conn_id=AWS_CONN,
            dag=self.dag,
            wait_for_completion=False,
        )
        _patch_hook_get_connection(delete_db_instance_operator.hook)
        delete_db_instance_operator.execute(None)

        with pytest.raises(self.hook.conn.exceptions.ClientError):
            self.hook.conn.describe_db_instances(DBInstanceIdentifier=DB_INSTANCE_NAME)
        mock_await_status.assert_not_called()

    def test_template_fields(self):
        operator = RdsDeleteDbInstanceOperator(
            task_id="test_delete_db_instance_no_wait",
            db_instance_identifier=DB_INSTANCE_NAME,
            rds_kwargs={
                "SkipFinalSnapshot": True,
            },
            aws_conn_id=AWS_CONN,
            wait_for_completion=False,
            region_name=REGION,
        )
        validate_template_fields(operator)


class TestRdsStopDbOperator:
    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.hook = RdsHook(aws_conn_id=AWS_CONN, region_name="us-east-1")
        _patch_hook_get_connection(cls.hook)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.hook

    @mock_aws
    def test_stop_db_instance(self):
        _create_db_instance(self.hook)
        stop_db_instance = RdsStopDbOperator(task_id="test_stop_db_instance", db_identifier=DB_INSTANCE_NAME)
        _patch_hook_get_connection(stop_db_instance.hook)
        stop_db_instance.execute(None)
        result = self.hook.conn.describe_db_instances(DBInstanceIdentifier=DB_INSTANCE_NAME)
        status = result["DBInstances"][0]["DBInstanceStatus"]
        assert status == "stopped"

    @mock_aws
    @patch.object(RdsHook, "get_waiter")
    def test_stop_db_instance_no_wait(self, mock_get_waiter):
        _create_db_instance(self.hook)
        stop_db_instance = RdsStopDbOperator(
            task_id="test_stop_db_instance_no_wait", db_identifier=DB_INSTANCE_NAME, wait_for_completion=False
        )
        _patch_hook_get_connection(stop_db_instance.hook)
        stop_db_instance.execute(None)
        result = self.hook.conn.describe_db_instances(DBInstanceIdentifier=DB_INSTANCE_NAME)
        status = result["DBInstances"][0]["DBInstanceStatus"]
        assert status == "stopped"
        mock_get_waiter.assert_not_called()

    @mock.patch.object(RdsHook, "conn")
    def test_deferred(self, conn_mock):
        op = RdsStopDbOperator(
            task_id="test_stop_db_instance_no_wait",
            db_identifier=DB_INSTANCE_NAME,
            deferrable=True,
        )

        with pytest.raises(TaskDeferred) as defer:
            op.execute({})

        assert isinstance(defer.value.trigger, RdsDbStoppedTrigger)

    @mock_aws
    def test_stop_db_instance_create_snapshot(self):
        _create_db_instance(self.hook)
        stop_db_instance = RdsStopDbOperator(
            task_id="test_stop_db_instance_create_snapshot",
            db_identifier=DB_INSTANCE_NAME,
            db_snapshot_identifier=DB_INSTANCE_SNAPSHOT,
        )
        _patch_hook_get_connection(stop_db_instance.hook)
        stop_db_instance.execute(None)

        describe_result = self.hook.conn.describe_db_instances(DBInstanceIdentifier=DB_INSTANCE_NAME)
        status = describe_result["DBInstances"][0]["DBInstanceStatus"]
        assert status == "stopped"

        snapshot_result = self.hook.conn.describe_db_snapshots(DBSnapshotIdentifier=DB_INSTANCE_SNAPSHOT)
        instance_snapshots = snapshot_result.get("DBSnapshots")
        assert instance_snapshots
        assert len(instance_snapshots) == 1

    @mock_aws
    def test_stop_db_cluster(self):
        _create_db_cluster(self.hook)
        stop_db_cluster = RdsStopDbOperator(
            task_id="test_stop_db_cluster", db_identifier=DB_CLUSTER_NAME, db_type="cluster"
        )
        _patch_hook_get_connection(stop_db_cluster.hook)
        stop_db_cluster.execute(None)

        describe_result = self.hook.conn.describe_db_clusters(DBClusterIdentifier=DB_CLUSTER_NAME)
        status = describe_result["DBClusters"][0]["Status"]
        assert status == "stopped"

    @mock_aws
    def test_stop_db_cluster_create_snapshot_logs_warning_message(self, caplog):
        _create_db_cluster(self.hook)
        stop_db_cluster = RdsStopDbOperator(
            task_id="test_stop_db_cluster",
            db_identifier=DB_CLUSTER_NAME,
            db_type="cluster",
            db_snapshot_identifier=DB_CLUSTER_SNAPSHOT,
        )
        _patch_hook_get_connection(stop_db_cluster.hook)
        with caplog.at_level(logging.WARNING, logger=stop_db_cluster.log.name):
            stop_db_cluster.execute(None)
        warning_message = (
            "'db_snapshot_identifier' does not apply to db clusters. Remove it to silence this warning."
        )
        assert warning_message in caplog.text

    def test_template_fields(self):
        operator = RdsStopDbOperator(
            task_id="test_stop_db_cluster",
            db_identifier=DB_CLUSTER_NAME,
            db_type="cluster",
            db_snapshot_identifier=DB_CLUSTER_SNAPSHOT,
            region_name=REGION,
        )
        validate_template_fields(operator)


class TestRdsStartDbOperator:
    @classmethod
    def setup_class(cls):
        cls.dag = DAG(
            dag_id="test_dag",
            schedule=None,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
        )
        cls.hook = RdsHook(aws_conn_id=AWS_CONN, region_name="us-east-1")
        _patch_hook_get_connection(cls.hook)

    @classmethod
    def teardown_class(cls):
        del cls.dag
        del cls.hook

    @mock_aws
    def test_start_db_instance(self):
        _create_db_instance(self.hook)
        self.hook.conn.stop_db_instance(DBInstanceIdentifier=DB_INSTANCE_NAME)
        result_before = self.hook.conn.describe_db_instances(DBInstanceIdentifier=DB_INSTANCE_NAME)
        status_before = result_before["DBInstances"][0]["DBInstanceStatus"]
        assert status_before == "stopped"

        start_db_instance = RdsStartDbOperator(
            task_id="test_start_db_instance", db_identifier=DB_INSTANCE_NAME
        )
        _patch_hook_get_connection(start_db_instance.hook)
        start_db_instance.execute(None)

        result_after = self.hook.conn.describe_db_instances(DBInstanceIdentifier=DB_INSTANCE_NAME)
        status_after = result_after["DBInstances"][0]["DBInstanceStatus"]
        assert status_after == "available"

    @mock_aws
    def test_start_db_cluster(self):
        _create_db_cluster(self.hook)
        self.hook.conn.stop_db_cluster(DBClusterIdentifier=DB_CLUSTER_NAME)
        result_before = self.hook.conn.describe_db_clusters(DBClusterIdentifier=DB_CLUSTER_NAME)
        status_before = result_before["DBClusters"][0]["Status"]
        assert status_before == "stopped"

        start_db_cluster = RdsStartDbOperator(
            task_id="test_start_db_cluster", db_identifier=DB_CLUSTER_NAME, db_type="cluster"
        )
        _patch_hook_get_connection(start_db_cluster.hook)
        start_db_cluster.execute(None)

        result_after = self.hook.conn.describe_db_clusters(DBClusterIdentifier=DB_CLUSTER_NAME)
        status_after = result_after["DBClusters"][0]["Status"]
        assert status_after == "available"

    @mock.patch.object(RdsHook, "conn")
    def test_deferred(self, conn_mock):
        op = RdsStartDbOperator(
            task_id="test_stop_db_instance_no_wait",
            db_identifier=DB_INSTANCE_NAME,
            deferrable=True,
        )

        with pytest.raises(TaskDeferred) as defer:
            op.execute({})

        assert isinstance(defer.value.trigger, RdsDbAvailableTrigger)

    def test_template_fields(self):
        operator = RdsStartDbOperator(
            region_name=REGION,
            aws_conn_id=AWS_CONN,
            task_id="test_start_db_cluster",
            db_identifier=DB_CLUSTER_NAME,
            db_type="cluster",
        )
        validate_template_fields(operator)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Note:  DMS requires you to configure specific IAM roles/permissions.  For more information, see
https://docs.aws.amazon.com/dms/latest/userguide/security-iam.html#CHAP_Security.APIRole
"""

from __future__ import annotations

import json
from datetime import datetime

import boto3
from sqlalchemy import Column, MetaData, String, Table, create_engine

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.dms import (
    DmsCreateReplicationConfigOperator,
    DmsDeleteReplicationConfigOperator,
    DmsDescribeReplicationConfigsOperator,
    DmsDescribeReplicationsOperator,
)
from airflow.providers.amazon.aws.operators.rds import (
    RdsCreateDbInstanceOperator,
    RdsDeleteDbInstanceOperator,
)
from airflow.providers.amazon.aws.operators.s3 import S3CreateBucketOperator, S3DeleteBucketOperator
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

"""
This example demonstrates how to use the DMS operators to create a serverless replication task to replicate data
from a PostgreSQL database to Amazon S3.

The IAM role used for the replication must have the permissions defined in the [Amazon S3 target](https://docs.aws.amazon.com/dms/latest/userguide/CHAP_Target.S3.html#CHAP_Target.S3.Prerequisites)
documentation.
"""

DAG_ID = "example_dms_serverless"
ROLE_ARN_KEY = "ROLE_ARN"

sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

# Config values for setting up the "Source" database.
CA_CERT_ID = "rds-ca-rsa2048-g1"
RDS_ENGINE = "postgres"
RDS_PROTOCOL = "postgresql"
RDS_USERNAME = "username"
# NEVER store your production password in plaintext in a DAG like this.
# Use Airflow Secrets or a secret manager for this in production.
RDS_PASSWORD = "rds_password"
TABLE_HEADERS = ["apache_project", "release_year"]
SAMPLE_DATA = [
    ("Airflow", "2015"),
    ("OpenOffice", "2012"),
    ("Subversion", "2000"),
    ("NiFi", "2006"),
]


def _get_rds_instance_endpoint(instance_name: str):
    print("Retrieving RDS instance endpoint.")
    rds_client = boto3.client("rds")

    response = rds_client.describe_db_instances(DBInstanceIdentifier=instance_name)
    rds_instance_endpoint = response["DBInstances"][0]["Endpoint"]
    return rds_instance_endpoint


@task
def create_sample_table(instance_name: str, db_name: str, table_name: str):
    print("Creating sample table.")

    rds_endpoint = _get_rds_instance_endpoint(instance_name)
    hostname = rds_endpoint["Address"]
    port = rds_endpoint["Port"]
    rds_url = f"{RDS_PROTOCOL}://{RDS_USERNAME}:{RDS_PASSWORD}@{hostname}:{port}/{db_name}"
    engine = create_engine(rds_url)

    table = Table(
        table_name,
        MetaData(engine),
        Column(TABLE_HEADERS[0], String, primary_key=True),
        Column(TABLE_HEADERS[1], String),
    )

    with engine.connect() as connection:
        # Create the Table.
        table.create()
        load_data = table.insert().values(SAMPLE_DATA)
        connection.execute(load_data)

        # Read the data back to verify everything is working.
        connection.execute(table.select())


@task(multiple_outputs=True)
def create_dms_assets(
    db_name: str,
    instance_name: str,
    bucket_name: str,
    role_arn,
    source_endpoint_identifier: str,
    target_endpoint_identifier: str,
    table_definition: dict,
):
    print("Creating DMS assets.")
    dms_client = boto3.client("dms")
    rds_instance_endpoint = _get_rds_instance_endpoint(instance_name)

    print("Creating DMS source endpoint.")
    source_endpoint_arn = dms_client.create_endpoint(
        EndpointIdentifier=source_endpoint_identifier,
        EndpointType="source",
        EngineName=RDS_ENGINE,
        Username=RDS_USERNAME,
        Password=RDS_PASSWORD,
        ServerName=rds_instance_endpoint["Address"],
        Port=rds_instance_endpoint["Port"],
        DatabaseName=db_name,
        SslMode="require",
    )["Endpoint"]["EndpointArn"]

    print("Creating DMS target endpoint.")
    target_endpoint_arn = dms_client.create_endpoint(
        EndpointIdentifier=target_endpoint_identifier,
        EndpointType="target",
        EngineName="s3",
        S3Settings={
            "BucketName": bucket_name,
            "BucketFolder": "folder",
            "ServiceAccessRoleArn": role_arn,
            "ExternalTableDefinition": json.dumps(table_definition),
        },
    )["Endpoint"]["EndpointArn"]

    return {
        "source_endpoint_arn": source_endpoint_arn,
        "target_endpoint_arn": target_endpoint_arn,
    }


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_dms_assets(
    source_endpoint_arn: str,
    target_endpoint_arn: str,
    source_endpoint_identifier: str,
    target_endpoint_identifier: str,
):
    dms_client = boto3.client("dms")
    dms_client.delete_endpoint(EndpointArn=source_endpoint_arn)
    dms_client.delete_endpoint(EndpointArn=target_endpoint_arn)
    dms_client.get_waiter("endpoint_deleted").wait(
        Filters=[
            {
                "Name": "endpoint-id",
                "Values": [source_endpoint_identifier, target_endpoint_identifier],
            }
        ]
    )


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]
    role_arn = test_context[ROLE_ARN_KEY]

    bucket_name = f"{env_id}-dms-serverless-bucket"
    rds_instance_name = f"{env_id}-instance"
    rds_db_name = f"{env_id}_source_database"  # dashes are not allowed in db name
    rds_table_name = f"{env_id}-table"
    source_endpoint_identifier = f"{env_id}-source-endpoint"
    target_endpoint_identifier = f"{env_id}-target-endpoint"
    replication_id = f"{env_id}-replication-id"

    create_s3_bucket = S3CreateBucketOperator(task_id="create_s3_bucket", bucket_name=bucket_name)

    create_db_instance = RdsCreateDbInstanceOperator(
        task_id="create_db_instance",
        db_instance_identifier=rds_instance_name,
        db_instance_class="db.t3.micro",
        engine=RDS_ENGINE,
        rds_kwargs={
            "DBName": rds_db_name,
            "AllocatedStorage": 20,
            "MasterUsername": RDS_USERNAME,
            "MasterUserPassword": RDS_PASSWORD,
            "PubliclyAccessible": True,
        },
    )

    # Sample data.
    table_definition = {
        "TableCount": "1",
        "Tables": [
            {
                "TableName": rds_table_name,
                "TableColumns": [
                    {
                        "ColumnName": TABLE_HEADERS[0],
                        "ColumnType": "STRING",
                        "ColumnNullable": "false",
                        "ColumnIsPk": "true",
                    },
                    {"ColumnName": TABLE_HEADERS[1], "ColumnType": "STRING", "ColumnLength": "4"},
                ],
                "TableColumnsTotal": "2",
            }
        ],
    }
    table_mappings = {
        "rules": [
            {
                "rule-type": "selection",
                "rule-id": "1",
                "rule-name": "1",
                "object-locator": {
                    "schema-name": "public",
                    "table-name": rds_table_name,
                },
                "rule-action": "include",
            }
        ]
    }

    create_assets = create_dms_assets(
        db_name=rds_db_name,
        instance_name=rds_instance_name,
        bucket_name=bucket_name,
        role_arn=role_arn,
        source_endpoint_identifier=source_endpoint_identifier,
        target_endpoint_identifier=target_endpoint_identifier,
        table_definition=table_definition,
    )

    # [START howto_operator_dms_create_replication_config]
    create_replication_config = DmsCreateReplicationConfigOperator(
        task_id="create_replication_config",
        replication_config_id=replication_id,
        source_endpoint_arn=create_assets["source_endpoint_arn"],
        target_endpoint_arn=create_assets["target_endpoint_arn"],
        compute_config={
            "MaxCapacityUnits": 4,
            "MinCapacityUnits": 1,
            "MultiAZ": False,
            "ReplicationSubnetGroupId": "default",
        },
        replication_type="full-load",
        table_mappings=json.dumps(table_mappings),
    )
    # [END howto_operator_dms_create_replication_config]

    # [START howto_operator_dms_describe_replication_config]
    describe_replication_configs = DmsDescribeReplicationConfigsOperator(
        task_id="describe_replication_configs",
    )
    # [END howto_operator_dms_describe_replication_config]

    # [START howto_operator_dms_serverless_describe_replication]
    describe_replications = DmsDescribeReplicationsOperator(
        task_id="describe_replications",
    )
    # [END howto_operator_dms_serverless_describe_replication]

    # Comment the next two tasks because they take too much time to be run in the CI
    # Keep them for documentation purposes
    """
    # [START howto_operator_dms_serverless_start_replication]
    replicate = DmsStartReplicationOperator(
        task_id="replicate",
        replication_config_arn="{{ task_instance.xcom_pull(task_ids='create_replication_config', key='return_value') }}",
        replication_start_type="start-replication",
        wait_for_completion=True,
        waiter_delay=60,
        waiter_max_attempts=200,
    )
    # [END howto_operator_dms_serverless_start_replication]

    # [START howto_operator_dms_serverless_stop_replication]
    stop_replication = DmsStopReplicationOperator(
        task_id="stop_replication",
        replication_config_arn="{{ task_instance.xcom_pull(task_ids='create_replication_config', key='return_value') }}",
        wait_for_completion=True,
        waiter_delay=120,
        waiter_max_attempts=200,
    )
    # [END howto_operator_dms_serverless_stop_replication]
    """

    # [START howto_operator_dms_serverless_delete_replication_config]
    delete_replication_config = DmsDeleteReplicationConfigOperator(
        task_id="delete_replication_config",
        waiter_max_attempts=200,
        replication_config_arn="{{ task_instance.xcom_pull(task_ids='create_replication_config', key='return_value') }}",
    )
    # [END howto_operator_dms_serverless_delete_replication_config]
    delete_replication_config.trigger_rule = TriggerRule.ALL_DONE

    delete_assets = delete_dms_assets(
        source_endpoint_arn=create_assets["source_endpoint_arn"],
        target_endpoint_arn=create_assets["target_endpoint_arn"],
        source_endpoint_identifier=source_endpoint_identifier,
        target_endpoint_identifier=target_endpoint_identifier,
    )

    delete_db_instance = RdsDeleteDbInstanceOperator(
        task_id="delete_db_instance",
        db_instance_identifier=rds_instance_name,
        rds_kwargs={
            "SkipFinalSnapshot": True,
        },
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_s3_bucket = S3DeleteBucketOperator(
        task_id="delete_s3_bucket",
        bucket_name=bucket_name,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        test_context,
        create_s3_bucket,
        create_db_instance,
        create_sample_table(rds_instance_name, rds_db_name, rds_table_name),
        create_assets,
        # TEST BODY
        create_replication_config,
        describe_replication_configs,
        describe_replications,
        delete_replication_config,
        # TEST TEARDOWN
        delete_assets,
        delete_db_instance,
        delete_s3_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

import boto3

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.sns import SnsPublishOperator
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import SystemTestContextBuilder

sys_test_context_task = SystemTestContextBuilder().build()

DAG_ID = "example_sns"


@task
def create_topic(topic_name) -> str:
    return boto3.client("sns").create_topic(Name=topic_name)["TopicArn"]


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_topic(topic_arn) -> None:
    boto3.client("sns").delete_topic(TopicArn=topic_arn)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]

    sns_topic_name = f"{env_id}-test-topic"

    create_sns_topic = create_topic(sns_topic_name)

    # [START howto_operator_sns_publish_operator]
    publish_message = SnsPublishOperator(
        task_id="publish_message",
        target_arn=create_sns_topic,
        message="This is a sample message sent to SNS via an Apache Airflow DAG task.",
    )
    # [END howto_operator_sns_publish_operator]

    chain(
        # TEST SETUP
        test_context,
        create_sns_topic,
        # TEST BODY
        publish_message,
        # TEST TEARDOWN
        delete_topic(create_sns_topic),
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from datetime import datetime
from os import environ

import boto3

from airflow.decorators import task, task_group
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.hooks.bedrock import BedrockHook
from airflow.providers.amazon.aws.operators.bedrock import (
    BedrockCreateProvisionedModelThroughputOperator,
    BedrockCustomizeModelOperator,
    BedrockInvokeModelOperator,
)
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.sensors.bedrock import (
    BedrockCustomizeModelCompletedSensor,
    BedrockProvisionModelThroughputCompletedSensor,
)
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import SystemTestContextBuilder

# Externally fetched variables:
ROLE_ARN_KEY = "ROLE_ARN"
sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

DAG_ID = "example_bedrock"

# Creating a custom model takes nearly two hours. If SKIP_LONG_TASKS
# is True then these tasks will be skipped. This way we can still have
# the code snippets for docs, and we can manually run the full tests.
SKIP_LONG_TASKS = environ.get("SKIP_LONG_SYSTEM_TEST_TASKS", default=True)

# No-commitment Provisioned Throughput is currently restricted to external
# customers only and will fail with a ServiceQuotaExceededException if run
# on the AWS System Test stack.
SKIP_PROVISION_THROUGHPUT = environ.get("SKIP_RESTRICTED_SYSTEM_TEST_TASKS", default=True)


LLAMA_SHORT_MODEL_ID = "meta.llama3-8b-instruct-v1:0"
TITAN_MODEL_ID = "amazon.titan-text-express-v1:0:8k"
TITAN_SHORT_MODEL_ID = TITAN_MODEL_ID.split(":")[0]

PROMPT = "What color is an orange?"
TRAIN_DATA = {"prompt": "what is AWS", "completion": "it's Amazon Web Services"}
HYPERPARAMETERS = {
    "epochCount": "1",
    "batchSize": "1",
    "learningRate": ".0005",
    "learningRateWarmupSteps": "0",
}


@task_group
def customize_model_workflow():
    # [START howto_operator_customize_model]
    customize_model = BedrockCustomizeModelOperator(
        task_id="customize_model",
        job_name=custom_model_job_name,
        custom_model_name=custom_model_name,
        role_arn=test_context[ROLE_ARN_KEY],
        base_model_id=f"{model_arn_prefix}{TITAN_SHORT_MODEL_ID}",
        hyperparameters=HYPERPARAMETERS,
        training_data_uri=training_data_uri,
        output_data_uri=f"s3://{bucket_name}/myOutputData",
    )
    # [END howto_operator_customize_model]

    # [START howto_sensor_customize_model]
    await_custom_model_job = BedrockCustomizeModelCompletedSensor(
        task_id="await_custom_model_job",
        job_name=custom_model_job_name,
    )
    # [END howto_sensor_customize_model]

    @task
    def delete_custom_model():
        BedrockHook().conn.delete_custom_model(modelIdentifier=custom_model_name)

    @task.short_circuit()
    def should_run_long_tasks():
        return not SKIP_LONG_TASKS

    chain(should_run_long_tasks(), customize_model, await_custom_model_job, delete_custom_model())


@task_group
def provision_throughput_workflow():
    # [START howto_operator_provision_throughput]
    provision_throughput = BedrockCreateProvisionedModelThroughputOperator(
        task_id="provision_throughput",
        model_units=1,
        provisioned_model_name=provisioned_model_name,
        model_id=f"{model_arn_prefix}{TITAN_MODEL_ID}",
    )
    # [END howto_operator_provision_throughput]

    # [START howto_sensor_provision_throughput]
    await_provision_throughput = BedrockProvisionModelThroughputCompletedSensor(
        task_id="await_provision_throughput",
        model_id=provision_throughput.output,
    )
    # [END howto_sensor_provision_throughput]

    @task
    def delete_provision_throughput(provisioned_model_id: str):
        BedrockHook().conn.delete_provisioned_model_throughput(provisionedModelId=provisioned_model_id)

    @task.short_circuit()
    def should_run_provision_throughput():
        return not SKIP_PROVISION_THROUGHPUT

    chain(
        should_run_provision_throughput(),
        provision_throughput,
        await_provision_throughput,
        delete_provision_throughput(provision_throughput.output),
    )


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]
    bucket_name = f"{env_id}-bedrock"
    input_data_s3_key = f"{env_id}/train.jsonl"
    training_data_uri = f"s3://{bucket_name}/{input_data_s3_key}"
    custom_model_name = f"CustomModel{env_id}"
    custom_model_job_name = f"CustomizeModelJob{env_id}"
    provisioned_model_name = f"ProvisionedModel{env_id}"
    model_arn_prefix = f"arn:aws:bedrock:{boto3.session.Session().region_name}::foundation-model/"

    create_bucket = S3CreateBucketOperator(
        task_id="create_bucket",
        bucket_name=bucket_name,
    )

    upload_training_data = S3CreateObjectOperator(
        task_id="upload_data",
        s3_bucket=bucket_name,
        s3_key=input_data_s3_key,
        data=json.dumps(TRAIN_DATA),
    )

    # [START howto_operator_invoke_llama_model]
    invoke_llama_model = BedrockInvokeModelOperator(
        task_id="invoke_llama",
        model_id=LLAMA_SHORT_MODEL_ID,
        input_data={"prompt": PROMPT},
    )
    # [END howto_operator_invoke_llama_model]

    # [START howto_operator_invoke_titan_model]
    invoke_titan_model = BedrockInvokeModelOperator(
        task_id="invoke_titan",
        model_id=TITAN_SHORT_MODEL_ID,
        input_data={"inputText": PROMPT},
    )
    # [END howto_operator_invoke_titan_model]

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=bucket_name,
        force_delete=True,
    )

    chain(
        # TEST SETUP
        test_context,
        create_bucket,
        upload_training_data,
        # TEST BODY
        [invoke_llama_model, invoke_titan_model],
        customize_model_workflow(),
        provision_throughput_workflow(),
        # TEST TEARDOWN
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from datetime import datetime

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.hooks.redshift_cluster import RedshiftHook
from airflow.providers.amazon.aws.operators.redshift_cluster import (
    RedshiftCreateClusterOperator,
    RedshiftDeleteClusterOperator,
)
from airflow.providers.amazon.aws.operators.redshift_data import RedshiftDataOperator
from airflow.providers.amazon.aws.operators.s3 import S3CreateBucketOperator, S3DeleteBucketOperator
from airflow.providers.amazon.aws.sensors.redshift_cluster import RedshiftClusterSensor
from airflow.providers.amazon.aws.transfers.sql_to_s3 import SqlToS3Operator
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder
from tests_common.test_utils.api_client_helpers import make_authenticated_rest_api_request

DAG_ID = "example_sql_to_s3"

# Externally fetched variables:
SECURITY_GROUP_KEY = "SECURITY_GROUP"
CLUSTER_SUBNET_GROUP_KEY = "CLUSTER_SUBNET_GROUP"

sys_test_context_task = (
    SystemTestContextBuilder().add_variable(SECURITY_GROUP_KEY).add_variable(CLUSTER_SUBNET_GROUP_KEY).build()
)


DB_LOGIN = "adminuser"
DB_PASS = "MyAmazonPassword1"
DB_NAME = "dev"

REDSHIFT_TABLE = "test_table"
SQL_QUERY = f"SELECT * FROM {REDSHIFT_TABLE}"

SQL_CREATE_TABLE = f"""
    CREATE TABLE IF NOT EXISTS {REDSHIFT_TABLE} (
    fruit_id INTEGER,
    name VARCHAR NOT NULL,
    color VARCHAR NOT NULL
    );
"""

SQL_INSERT_DATA = f"INSERT INTO {REDSHIFT_TABLE} VALUES ( 1, 'Banana', 'Yellow');"


@task
def create_connection(conn_id_name: str, cluster_id: str):
    redshift_hook = RedshiftHook()
    cluster_endpoint = redshift_hook.get_conn().describe_clusters(ClusterIdentifier=cluster_id)["Clusters"][0]
    make_authenticated_rest_api_request(
        path="/api/v2/connections",
        method="POST",
        body={
            "connection_id": conn_id_name,
            "conn_type": "redshift",
            "host": cluster_endpoint["Endpoint"]["Address"],
            "login": DB_LOGIN,
            "schema": cluster_endpoint["DBName"],
            "port": cluster_endpoint["Endpoint"]["Port"],
            "password": DB_PASS,
        },
    )


with DAG(
    dag_id=DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule="@once",
    catchup=False,
    tags=["example"],
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]
    security_group_id = test_context[SECURITY_GROUP_KEY]
    cluster_subnet_group_name = test_context[CLUSTER_SUBNET_GROUP_KEY]
    redshift_cluster_identifier = f"{env_id}-redshift-cluster"
    conn_id_name = f"{env_id}-conn-id"
    sg_name = f"{env_id}-sg"
    bucket_name = f"{env_id}-s3-bucket"
    key = f"{env_id}-key"

    create_bucket = S3CreateBucketOperator(
        task_id="create_bucket",
        bucket_name=bucket_name,
    )

    create_cluster = RedshiftCreateClusterOperator(
        task_id="create_cluster",
        cluster_identifier=redshift_cluster_identifier,
        vpc_security_group_ids=[security_group_id],
        cluster_subnet_group_name=cluster_subnet_group_name,
        cluster_type="single-node",
        node_type="dc2.large",
        master_username=DB_LOGIN,
        master_user_password=DB_PASS,
    )

    wait_cluster_available = RedshiftClusterSensor(
        task_id="wait_cluster_available",
        cluster_identifier=redshift_cluster_identifier,
        target_status="available",
        poke_interval=15,
        timeout=60 * 30,
    )

    set_up_connection = create_connection(conn_id_name, cluster_id=redshift_cluster_identifier)

    create_table_redshift_data = RedshiftDataOperator(
        task_id="create_table_redshift_data",
        cluster_identifier=redshift_cluster_identifier,
        database=DB_NAME,
        db_user=DB_LOGIN,
        sql=SQL_CREATE_TABLE,
        wait_for_completion=True,
    )

    insert_data = RedshiftDataOperator(
        task_id="insert_data",
        cluster_identifier=redshift_cluster_identifier,
        database=DB_NAME,
        db_user=DB_LOGIN,
        sql=SQL_INSERT_DATA,
        wait_for_completion=True,
    )

    # [START howto_transfer_sql_to_s3]
    sql_to_s3_task = SqlToS3Operator(
        task_id="sql_to_s3_task",
        sql_conn_id=conn_id_name,
        query=SQL_QUERY,
        s3_bucket=bucket_name,
        s3_key=key,
        replace=True,
    )
    # [END howto_transfer_sql_to_s3]

    # [START howto_transfer_sql_to_s3_with_groupby_param]
    sql_to_s3_task_with_groupby = SqlToS3Operator(
        task_id="sql_to_s3_with_groupby_task",
        sql_conn_id=conn_id_name,
        query=SQL_QUERY,
        s3_bucket=bucket_name,
        s3_key=key,
        replace=True,
        groupby_kwargs={"by": "color"},
    )
    # [END howto_transfer_sql_to_s3_with_groupby_param]

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=bucket_name,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_cluster = RedshiftDeleteClusterOperator(
        task_id="delete_cluster",
        cluster_identifier=redshift_cluster_identifier,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        test_context,
        create_bucket,
        create_cluster,
        wait_cluster_available,
        set_up_connection,
        create_table_redshift_data,
        insert_data,
        # TEST BODY
        sql_to_s3_task,
        sql_to_s3_task_with_groupby,
        # TEST TEARDOWN
        delete_bucket,
        delete_cluster,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.opensearch_serverless import OpenSearchServerlessHook
from airflow.providers.amazon.aws.sensors.opensearch_serverless import (
    OpenSearchServerlessCollectionActiveSensor,
)


class TestOpenSearchServerlessCollectionActiveSensor:
    def setup_method(self):
        self.default_op_kwargs = dict(
            task_id="test_sensor",
            collection_id="knowledge_base_id",
            poke_interval=5,
            max_retries=1,
        )
        self.sensor = OpenSearchServerlessCollectionActiveSensor(**self.default_op_kwargs, aws_conn_id=None)

    def test_base_aws_op_attributes(self):
        op = OpenSearchServerlessCollectionActiveSensor(**self.default_op_kwargs)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

        op = OpenSearchServerlessCollectionActiveSensor(
            **self.default_op_kwargs,
            aws_conn_id="aws-test-custom-conn",
            region_name="eu-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.aws_conn_id == "aws-test-custom-conn"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

    @pytest.mark.parametrize(
        "collection_name, collection_id, expected_pass",
        [
            pytest.param("name", "id", False, id="both_provided_fails"),
            pytest.param("name", None, True, id="only_name_provided_passes"),
            pytest.param(None, "id", True, id="only_id_provided_passes"),
        ],
    )
    def test_name_and_id_combinations(self, collection_name, collection_id, expected_pass):
        call_args = {
            "task_id": "test_sensor",
            "collection_name": collection_name,
            "collection_id": collection_id,
            "poke_interval": 5,
            "max_retries": 1,
        }
        if expected_pass:
            op = OpenSearchServerlessCollectionActiveSensor(**call_args)
            assert op.collection_id == collection_id
            assert op.collection_name == collection_name
        if not expected_pass:
            with pytest.raises(
                AttributeError, match="Either collection_ids or collection_names must be provided, not both."
            ):
                OpenSearchServerlessCollectionActiveSensor(**call_args)

    @pytest.mark.parametrize("state", list(OpenSearchServerlessCollectionActiveSensor.SUCCESS_STATES))
    @mock.patch.object(OpenSearchServerlessHook, "conn")
    def test_poke_success_states(self, mock_conn, state):
        mock_conn.batch_get_collection.return_value = {"collectionDetails": [{"status": state}]}
        assert self.sensor.poke({}) is True

    @pytest.mark.parametrize("state", list(OpenSearchServerlessCollectionActiveSensor.INTERMEDIATE_STATES))
    @mock.patch.object(OpenSearchServerlessHook, "conn")
    def test_poke_intermediate_states(self, mock_conn, state):
        mock_conn.batch_get_collection.return_value = {"collectionDetails": [{"status": state}]}
        assert self.sensor.poke({}) is False

    @pytest.mark.parametrize("state", list(OpenSearchServerlessCollectionActiveSensor.FAILURE_STATES))
    @mock.patch.object(OpenSearchServerlessHook, "conn")
    def test_poke_failure_states(self, mock_conn, state):
        mock_conn.batch_get_collection.return_value = {"collectionDetails": [{"status": state}]}
        sensor = OpenSearchServerlessCollectionActiveSensor(**self.default_op_kwargs, aws_conn_id=None)
        with pytest.raises(AirflowException, match=sensor.FAILURE_MESSAGE):
            sensor.poke({})
